# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Pydantic Model for WUFI-XML file format."""

from __future__ import annotations
from typing import Dict, Union, Any, Optional
from ph_units.converter import convert


# ------------------------------------------------------------------------------
# -- Unit Types


class BaseConverter:
    """Base Class for any types which return a value, converted to the right unit."""

    __unit_type__ = ""
    __value_type__ = int

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[str, Dict[str, Any]]):
        """
        Since WUFI XMl files come in several shapes and the XML tag may, or
        may not, include a "unit=__" attribute, we don't know what sort of input
        we're getting. So if it is a dict input, that means it should have a
        "unit" type. But it if it is a bare int or float, then we assume it is
        the correct unit already (SI) and just return it.
        """

        # -- If it is not a dict or None, just return the value
        if not isinstance(v, dict):
            return cls.__value_type__(v)

        # -- If its None, Return None
        if not v["value"]:
            return None

        # -- Otherwise, pull the value out of the dict and convert the value to the right unit
        try:
            result = convert(v["value"], v["unit_type"], cls.__unit_type__)
        except Exception:
            msg = f"Error converting input for '{cls.__name__}' using inputs: {v}"
            raise Exception(msg)
        if result is None:
            msg = (
                f"Error, could not convert:\n"
                f"\t{v['value']} from {v['unit_type']} to {cls.__unit_type__}"
            )
            raise Exception(msg)
        return cls.__value_type__(result)


class BaseCaster:
    """Base Class for any types which return a value, cast to the right type."""

    __unit_type__ = ""
    __value_type__ = int

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if isinstance(v, dict):
            if v["value"] is None:
                return None
            return cls.__value_type__(v["value"])
        elif v is None:
            return None
        else:
            return cls.__value_type__(v)


# ------------------------------------------------------------------------------
# -- Generics


class _Int(int, BaseCaster):
    __unit_type__ = ""
    __value_type__ = int


class _Float(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


class _Percentage(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Power


class Watts(float, BaseConverter):
    __unit_type__ = "W"
    __value_type__ = float


class Watts_per_M2K(float, BaseConverter):
    __unit_type__ = "W/M2K"
    __value_type__ = float


class Watts_per_MK(float, BaseConverter):
    __unit_type__ = "W/MK"
    __value_type__ = float


class Watts_per_M2(float, BaseConverter):
    __unit_type__ = "W/M2"
    __value_type__ = float


class Watts_per_DegreeK(float, BaseConverter):
    __unit_type__ = "W/K"
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Energy


class kWh(float, BaseConverter):
    __unit_type__ = "KWH"
    __value_type__ = float


class kWh_per_M2(float, BaseConverter):
    __unit_type__ = "KWH/M2"
    __value_type__ = float


class kWh_per_kWh(float):
    __unit_type__ = "WUFI_MEW"
    __value_type__ = float

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[str, Dict[str, Any]]) -> Optional[float]:
        if not isinstance(v, dict):
            return cls.__value_type__(v)

        if not v["value"]:
            return None

        if v["unit_type"] == "-":
            return cls.__value_type__(v["value"])

        result = convert(v["value"], v["unit_type"], cls.__unit_type__)
        if result is None:
            raise Exception(
                f"Could not convert: {v['value']} from {v['unit_type']} to {cls.__unit_type__}"
            )
        return cls.__value_type__(result)


class Wh_per_M3(float, BaseConverter):
    __unit_type__ = "WH/M3"
    __value_type__ = float


class Wh_per_M2K(float, BaseConverter):
    __unit_type__ = "WH/M2K"
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Size


class M(float, BaseConverter):
    __unit_type__ = "M"
    __value_type__ = float


class M3(float, BaseConverter):
    __unit_type__ = "M3"
    __value_type__ = float


class M2(float, BaseConverter):
    __unit_type__ = "M2"
    __value_type__ = float


class Liter(float):
    __unit_type__ = "L"
    __value_type__ = float

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[str, Dict[str, Any]]) -> Optional[float]:
        if not isinstance(v, dict):
            return cls.__value_type__(v)

        if not v["value"]:
            return None

        if v["unit_type"] == "-":
            return cls.__value_type__(v["value"])

        result = convert(v["value"], v["unit_type"], cls.__unit_type__)
        if result is None:
            raise Exception(
                f"Could not convert: {v['value']} from {v['unit_type']} to {cls.__unit_type__}"
            )
        return cls.__value_type__(result)


# ------------------------------------------------------------------------------
# -- Temperature


class DegreeC(float, BaseConverter):
    __unit_type__ = "C"
    __value_type__ = float


class DegreeDeltaK(float):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[str, Dict[str, Any]]):
        if not isinstance(v, dict):
            return float(v)

        type = v["unit_type"]
        if type == "K":
            type = "DELTA-K"
        elif type == "C":
            type = "DELTA-C"
        elif type == "°F":
            type = "DELTA-F"

        result = convert(v["value"], type, "DELTA-C")
        if result is None:
            raise Exception(f"Could not convert: {v['value']} from {v['unit_type']} to C")
        return float(result)


# ------------------------------------------------------------------------------
# -- Speed


class M_per_Second(float, BaseConverter):
    __unit_type__ = "M/S"
    __value_type__ = float


class M_per_Day(float, BaseConverter):
    __unit_type__ = "M/DAY"
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Density


class KG_per_M3(float, BaseConverter):
    __unit_type__ = "KG/M3"
    __value_type__ = float


class MG_per_M3(float, BaseConverter):
    __unit_type__ = "MG/M3"
    __value_type__ = float


class Joule_per_KGK(float, BaseConverter):
    __unit_type__ = "J/KGK"
    __value_type__ = float


class PartsPerMillionByVolume(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Time


class Hour(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


class Days_per_Year(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


class Hours_per_Year(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Lighting


class Lux(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Airflow


class M3_per_Hour(float, BaseConverter):
    __unit_type__ = "M3/HR"
    __value_type__ = float


class M3_per_Hour_per_M2(float, BaseConverter):
    __unit_type__ = "M3/HRM2"
    __value_type__ = float


class ACH(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Geometry


class AngleDegree(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


class CardinalDegrees(float, BaseCaster):
    __unit_type__ = ""
    __value_type__ = float


# ------------------------------------------------------------------------------
# -- Moisture Resistance


class WUFI_Vapor_Resistance_Factor(float):
    __unit_type__ = "WUFI_MEW"
    __value_type__ = float

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Union[str, Dict[str, Any]]) -> Optional[float]:
        if not isinstance(v, dict):
            return cls.__value_type__(v)

        if not v["value"]:
            return None

        if v["unit_type"] == "-":
            return cls.__value_type__(v["value"])

        result = convert(v["value"], v["unit_type"], cls.__unit_type__)
        if result is None:
            raise Exception(
                f"Could not convert: {v['value']} from {v['unit_type']} to {cls.__unit_type__}"
            )
        return cls.__value_type__(result)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create a new PhxBuilding from Honeybee-Rooms"""

from typing import List, Union, Dict
from functools import partial

from honeybee import room, aperture, face

from honeybee_energy.construction import windowshade, window
from honeybee_energy.properties.room import RoomEnergyProperties
from honeybee_energy.properties.aperture import ApertureEnergyProperties
from honeybee_energy.properties.face import FaceEnergyProperties

from honeybee_ph import space
from honeybee_ph.properties.aperture import AperturePhProperties
from honeybee_ph.properties.room import RoomPhProperties
from honeybee_energy_ph.properties.load.people import PeoplePhProperties
from honeybee_energy_ph.properties.construction.window import (
    WindowConstructionPhProperties,
)
from honeybee_energy_ph.properties.construction.opaque import (
    OpaqueConstructionPhProperties,
)

from PHX.from_HBJSON import create_geometry
from PHX.from_HBJSON.create_rooms import create_room_from_space
from PHX.model import building, constructions, components
from PHX.model.utilization_patterns import (
    UtilizationPatternCollection_Ventilation,
    UtilizationPatternCollection_Occupancy,
    UtilizationPatternCollection_Lighting,
)
from PHX.model.enums.building import (
    ComponentExposureExterior,
    ComponentFaceOpacity,
    ComponentColor,
    ComponentFaceType,
    ThermalBridgeType,
    SpecificHeatCapacity,
)


def _hb_face_opacity_to_phx_enum(_hb_face: face.Face) -> ComponentFaceOpacity:
    mapping = {
        "Wall": ComponentFaceOpacity.OPAQUE,
        "RoofCeiling": ComponentFaceOpacity.OPAQUE,
        "Floor": ComponentFaceOpacity.OPAQUE,
        "AirBoundary": ComponentFaceOpacity.AIRBOUNDARY,
    }
    return mapping[str(_hb_face.type)]


def _hb_face_type_to_phx_enum(_hb_face: face.Face) -> ComponentFaceType:
    mapping = {
        "Wall": ComponentFaceType.WALL,
        "Floor": ComponentFaceType.FLOOR,
        "RoofCeiling": ComponentFaceType.ROOF_CEILING,
        "AirBoundary": ComponentFaceType.AIR_BOUNDARY,
    }
    return mapping[str(_hb_face.type)]


def _hb_ext_exposure_to_phx_enum(
    _hb_face: Union[face.Face, aperture.Aperture]
) -> ComponentExposureExterior:
    mapping = {
        "Outdoors": ComponentExposureExterior.EXTERIOR,
        "Ground": ComponentExposureExterior.GROUND,
        "Surface": ComponentExposureExterior.SURFACE,
        "Adiabatic": ComponentExposureExterior.SURFACE,
    }
    return mapping[str(_hb_face.boundary_condition)]


def _hb_int_color_to_phx_enum(_hb_face: face.Face) -> ComponentColor:
    mapping = {
        "Wall": {
            "Outdoors": ComponentColor.EXT_WALL_OUTER,
            "Surface": ComponentColor.INNER_WALL,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.INNER_WALL,
        },
        "RoofCeiling": {
            "Outdoors": ComponentColor.SLOPED_ROOF_OUTER,
            "Surface": ComponentColor.CEILING,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.CEILING,
        },
        "Floor": {
            "Outdoors": ComponentColor.FLOOR,
            "Surface": ComponentColor.FLOOR,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.FLOOR,
        },
    }
    return mapping[str(_hb_face.type)][str(_hb_face.boundary_condition)]


def _hb_ext_color_to_phx_enum(_hb_face: face.Face) -> ComponentColor:
    mapping = {
        "Wall": {
            "Outdoors": ComponentColor.EXT_WALL_OUTER,
            "Surface": ComponentColor.INNER_WALL,
            "Ground": ComponentColor.EXT_WALL_OUTER,
            "Adiabatic": ComponentColor.INNER_WALL,
        },
        "RoofCeiling": {
            "Outdoors": ComponentColor.SLOPED_ROOF_INNER,
            "Surface": ComponentColor.CEILING,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.CEILING,
        },
        "Floor": {
            "Outdoors": ComponentColor.FLOOR,
            "Surface": ComponentColor.FLOOR,
            "Ground": ComponentColor.SURFACE_GROUND_CONTACT,
            "Adiabatic": ComponentColor.FLOOR,
        },
    }
    return mapping[str(_hb_face.type)][str(_hb_face.boundary_condition)]


def _get_hb_aperture_construction(_hb_aperture) -> window.WindowConstruction:
    """Return the Construction of the Aperture."""
    if hasattr(_hb_aperture.construction, "window_construction"):
        # -- its a WindowConstructionShade, so pull out the normal construction
        return _hb_aperture.construction.window_construction
    else:
        return _hb_aperture.construction


def create_component_from_hb_aperture(
    _host_compo: components.PhxComponentOpaque,
    _hb_aperture: aperture.Aperture,
    _hb_room: room.Room,
    _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
    _tolerance: float = 0.001,
) -> building.PhxComponentAperture:
    """Create a new Transparent (window) Component based on a Honeybee Aperture.

    Arguments:
    ----------
        * _host (components.PhxComponentOpaque): The Aperture's parent opaque component.
        * _hb_aperture (aperture.Aperture): The Honeybee-Aperture to use as the source.
        * _hb_room (room.Room): The Honeybee-Room to use as the source.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * (building.PhxComponentAperture): A new Transparent (window) Component.
    """

    # -- Type Aliases
    hb_ap_prop_energy: ApertureEnergyProperties = _hb_aperture.properties.energy  # type: ignore
    hb_ap_prop_ph: AperturePhProperties = _hb_aperture.properties.ph  # type: ignore
    hb_room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore
    hb_ap_const = _get_hb_aperture_construction(hb_ap_prop_energy)

    # -- Create new Aperture
    phx_ap = components.PhxComponentAperture(_host=_host_compo)
    phx_ap.display_name = _hb_aperture.display_name
    phx_ap.exposure_interior = hb_room_prop_ph.id_num
    phx_ap.window_type = _window_type_dict[hb_ap_const.identifier]
    phx_ap.variant_type_name = hb_ap_prop_ph.variant_type
    phx_ap._install_depth = hb_ap_prop_ph.install_depth
    monthly_fac = getattr(hb_ap_prop_ph, "default_monthly_shading_correction_factor", 1.0)
    phx_ap.default_monthly_shading_correction_factor = monthly_fac

    # -- Create new Aperture Element (Sash)
    new_phx_ap_element = components.PhxApertureElement(_host=phx_ap)
    new_phx_ap_element.display_name = _hb_aperture.display_name
    new_phx_ap_element.polygon = (
        create_geometry.create_PhxPolygonRectangular_from_hb_Face(
            _hb_aperture, _tolerance
        )
    )

    # -- Set the shading object dimensions, if there are any
    if hb_ap_prop_ph.shading_dimensions:
        shading_dims = hb_ap_prop_ph.shading_dimensions
        new_phx_ap_element.shading_dimensions.h_hori = shading_dims.h_hori
        new_phx_ap_element.shading_dimensions.d_hori = shading_dims.d_hori
        new_phx_ap_element.shading_dimensions.o_reveal = shading_dims.o_reveal
        new_phx_ap_element.shading_dimensions.d_reveal = shading_dims.d_reveal
        new_phx_ap_element.shading_dimensions.o_over = shading_dims.o_over
        new_phx_ap_element.shading_dimensions.d_over = shading_dims.d_over
    new_phx_ap_element.winter_shading_factor = hb_ap_prop_ph.winter_shading_factor
    new_phx_ap_element.summer_shading_factor = hb_ap_prop_ph.summer_shading_factor

    phx_ap.add_elements((new_phx_ap_element,))

    return phx_ap


def create_components_from_hb_face(
    _hb_face: face.Face,
    _hb_room: room.Room,
    _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
    _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
    _tolerance: float = 0.001,
) -> components.PhxComponentOpaque:
    """Returns a new Opaque Component (and any child components) based on a Honeybee Face,

    Arguments:
    ----------
        * _hb_face (face.Face): The Honeybee-Face to use as the source.
        * _hb_room (room.Room)L The Honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _window_type_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * (components.PhxComponentOpaque): The new Opaque Component.
    """

    # Type Aliases
    hb_face_prop_energy: FaceEnergyProperties = _hb_face.properties.energy  # type: ignore
    hb_face_const = hb_face_prop_energy.construction
    if not hasattr(hb_face_const, "properties"):
        msg = f"Error: Face Construction {hb_face_const.display_name} is missing .properties?"
    hb_face_const_prop_ph: OpaqueConstructionPhProperties = hb_face_const.properties.ph  # type: ignore
    hb_room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore

    # -- Build the new PHX Opaque Component
    opaque_compo = components.PhxComponentOpaque()
    opaque_compo.display_name = _hb_face.display_name

    opaque_compo.assembly = _assembly_dict[hb_face_prop_energy.construction.identifier]
    opaque_compo.assembly_type_id_num = hb_face_const_prop_ph.id_num

    opaque_compo.face_type = _hb_face_type_to_phx_enum(_hb_face)
    opaque_compo.face_opacity = _hb_face_opacity_to_phx_enum(_hb_face)
    opaque_compo.exposure_exterior = _hb_ext_exposure_to_phx_enum(_hb_face)
    opaque_compo.exposure_interior = hb_room_prop_ph.id_num
    opaque_compo.color_interior = _hb_int_color_to_phx_enum(_hb_face)
    opaque_compo.color_exterior = _hb_ext_color_to_phx_enum(_hb_face)

    # -- Create Polygon
    phx_polygon = create_geometry.create_PhxPolygon_from_hb_Face(_hb_face)
    opaque_compo.add_polygons(phx_polygon)

    # -- Create Child Apertures, register the Aperture with the Parent Compo
    for hb_aperture in _hb_face.apertures:
        phx_compo_aperture = create_component_from_hb_aperture(
            opaque_compo, hb_aperture, _hb_room, _window_type_dict, _tolerance
        )
        phx_polygon.add_child_poly_id(phx_compo_aperture.polygon_ids)
        opaque_compo.add_aperture(phx_compo_aperture)

    return opaque_compo


def create_components_from_hb_room(
    _hb_room: room.Room,
    _assembly_dict: Dict[str, constructions.PhxConstructionOpaque],
    _window_type_dict: Dict[str, constructions.PhxConstructionWindow],
    _tolerance: float = 0.001,
) -> List[components.PhxComponentOpaque]:
    """Create new Opaque and Transparent PHX-Components based on Honeybee-Room Faces.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.
        * _assembly_dict (Dict[str, constructions.PhxConstructionOpaque]): The Assembly Type dict.
        * _assembly_dict (Dict[str, constructions.PhxConstructionWindow]): The Window Type dict.

    Returns:
    --------
        * (List[components.PhxComponentOpaque]): A list of the new opaque PhxComponents.
    """
    return [
        create_components_from_hb_face(
            hb_face, _hb_room, _assembly_dict, _window_type_dict, _tolerance
        )
        for hb_face in _hb_room
    ]


def set_zone_occupancy(_hb_room: room.Room, zone: building.PhxZone) -> building.PhxZone:
    """Set the Zone's Residential Occupancy values."""
    # -- Type Aliases
    room_energy_prop: RoomEnergyProperties = _hb_room.properties.energy  # type: ignore
    hbph_people_prop: PeoplePhProperties = room_energy_prop.people.properties.ph  # type: ignore

    zone.res_occupant_quantity = hbph_people_prop.number_people
    zone.res_number_bedrooms = hbph_people_prop.number_bedrooms
    zone.res_number_dwellings = hbph_people_prop.number_dwelling_units

    return zone


def create_zones_from_hb_room(
    _hb_room: room.Room,
    _vent_sched_collection: UtilizationPatternCollection_Ventilation,
    _occ_sched_collection: UtilizationPatternCollection_Occupancy,
    _lighting_sched_collection: UtilizationPatternCollection_Lighting,
) -> building.PhxZone:
    """Create a new PHX-Zone based on a honeybee-Room.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.
        * _vent_sched_collection (UtilizationPatternCollection_Ventilation): The Ventilation Schedule Collection.
        * _occ_sched_collection (UtilizationPatternCollection_Occupancy): The Occupancy Schedule Collection.

    Returns:
    --------
        * (building.PhxZone): The new PHX-Zone object.
    """
    new_zone = building.PhxZone()

    new_zone.id_num = building.PhxZone._count
    new_zone.display_name = _hb_room.display_name

    # -- Sort the HB-Room's Spaces by their full_name
    rm_prop_ph = _hb_room.properties.ph  # type: RoomPhProperties # type: ignore
    sorted_spaces = sorted(rm_prop_ph.spaces, key=lambda space: space.full_name)

    # -- Create a new WUFI-Space (Room) for each HBPH-Space
    _create_space = partial(
        create_room_from_space,
        _vent_sched_collection=_vent_sched_collection,
        _occ_sched_collection=_occ_sched_collection,
        _lighting_sched_collection=_lighting_sched_collection,
    )
    new_zone.spaces = [_create_space(sp) for sp in sorted_spaces]

    # -- Set Zone properties
    new_zone.volume_gross = _hb_room.volume
    new_zone.weighted_net_floor_area = sum(
        (rm.weighted_floor_area for rm in new_zone.spaces)
    )
    new_zone.volume_net = sum((rm.net_volume for rm in new_zone.spaces))
    new_zone.specific_heat_capacity = SpecificHeatCapacity(
        rm_prop_ph.specific_heat_capacity.number
    )

    # -- Set the Zone's Occupancy based on the merged HB room
    new_zone = set_zone_occupancy(_hb_room, new_zone)

    # -- Set the Zones' thermal bridges
    for phx_thermal_bridge in create_thermal_bridges_from_hb_room(_hb_room):
        new_zone.add_thermal_bridges(phx_thermal_bridge)

    return new_zone


def create_thermal_bridges_from_hb_room(
    _hb_room: room.Room,
) -> List[components.PhxComponentThermalBridge]:
    """Create a list of new PHX-ThermalBridges based on those found on a honeybee-Room.

    Arguments:
    ----------
        * _hb_room (room.Room): The honeybee-Room to use as the source.

    Returns:
    --------
        * (List[components.PhxComponentThermalBridge]): A list of the new Thermal Bridge objects.
    """

    # -- Type Aliases
    hb_room_prop_ph: RoomPhProperties = _hb_room.properties.ph  # type: ignore

    # -- Create all the new Thermal Bridges
    phx_thermal_bridges = []
    for thermal_bridge in sorted(
        hb_room_prop_ph.ph_bldg_segment.thermal_bridges.values(),
        key=lambda tb: tb.display_name,
    ):
        phx_tb = components.PhxComponentThermalBridge()
        phx_tb.display_name = thermal_bridge.display_name
        phx_tb.quantity = thermal_bridge.quantity
        phx_tb.group_number = ThermalBridgeType(thermal_bridge.group_type.number)
        phx_tb.identifier = str(thermal_bridge.identifier)
        phx_tb.psi_value = thermal_bridge.psi_value
        phx_tb.fRsi_value = thermal_bridge.fRsi_value
        phx_tb.length = thermal_bridge.length
        phx_thermal_bridges.append(phx_tb)

    return phx_thermal_bridges



from PHX.PHPP.sheet_io.io_verification import *
from PHX.PHPP.sheet_io.io_climate import *
from PHX.PHPP.sheet_io.io_areas import *
from PHX.PHPP.sheet_io.io_u_values import *
from PHX.PHPP.sheet_io.io_components import *
from PHX.PHPP.sheet_io.io_windows import *
from PHX.PHPP.sheet_io.io_shading import *
from PHX.PHPP.sheet_io.io_addnl_vent import *
from PHX.PHPP.sheet_io.io_ventilation import *
from PHX.PHPP.sheet_io.io_hot_water import *
from PHX.PHPP.sheet_io.io_electricity import *
from PHX.PHPP.sheet_io.io_variants import *
from PHX.PHPP.sheet_io.io_PER import *
from PHX.PHPP.sheet_io.io_heating_demand import *
from PHX.PHPP.sheet_io.io_heating_peak_load import *
from PHX.PHPP.sheet_io.io_cooling_demand import *
from PHX.PHPP.sheet_io.io_cooling_peak_load import *
from PHX.PHPP.sheet_io.io_overview import *
from PHX.PHPP.sheet_io.io_elec_non_res import *
from PHX.PHPP.sheet_io.io_use_non_res import *
from PHX.PHPP.sheet_io.io_ihg_non_res import *
from PHX.PHPP.sheet_io.io_cooling_units import *
from PHX.PHPP.sheet_io.io_solar_dhw import *
from PHX.PHPP.sheet_io.io_solar_pv import *



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Ventilation Ducting Distribution Objects."""


from __future__ import annotations
from typing import Dict, List, Optional, ClassVar
from dataclasses import dataclass, field

from ladybug_geometry.geometry3d.polyline import LineSegment3D

from PHX.model.enums.hvac import PhxVentDuctType

@dataclass
class PhxDuctSegment:
    """An individual Duct Segment Segment."""

    identifier: str
    display_name: str
    geometry: LineSegment3D
    diameter: float
    height: Optional[float]
    width: Optional[float]
    insulation_thickness: float
    insulation_conductivity: float
    insulation_reflective: bool

    @property
    def length(self) -> float:
        return self.geometry.length
    
@dataclass
class PhxDuctElement:
    """A Duct Element / Run made of one or more PhxDuctSegments."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    identifier: str
    display_name: str
    duct_type: PhxVentDuctType = field(init=False, default=PhxVentDuctType.SUPPLY)
    vent_unit_id: int
    _segments: Dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def quantity(self) -> int:
        return 1

    @property
    def segments(self) -> List[PhxDuctSegment]:
        return list(self._segments.values())

    @property
    def length(self) -> float:
        return sum(_.length for _ in self.segments)

    @property
    def diameter(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.diameter
        return weighted_total / self.length

    @property
    def height(self) -> float: 
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * (seg.height or 0.0)
        return weighted_total / self.length 
    
    @property
    def width(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * (seg.width or 0.0)
        return weighted_total / self.length 

    @property
    def insulation_thickness(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.insulation_thickness
        return weighted_total / self.length
    
    @property
    def insulation_conductivity(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.insulation_conductivity
        return weighted_total / self.length
    
    @property
    def duct_shape(self) -> int:
        if self.height and self.width:
            return 2 # Rectangular Duct
        else:
            return 1 # Round Duct

    @property
    def is_reflective(self) -> bool:
        return any({seg.insulation_reflective for seg in self.segments})

    @property
    def assigned_vent_unit_ids(self) -> List[int]:
        return [self.vent_unit_id]

    def add_segment(self, _s: PhxDuctSegment) -> None:
        self._segments[_s.identifier] = _s



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Geometry Classes"""

from __future__ import annotations
from dataclasses import dataclass, field
import math
from typing import ClassVar, Collection, List, Union, Optional


class PolygonEdgeError(Exception):
    def __init__(self, _polygon: PhxPolygonRectangular, _segment_name: str):
        self.msg = (
            f"Error: Cannot create PhxPolygonRectangle {_polygon}"
            f"segment {_segment_name}. Missing 1 or more vertices?"
        )
        super().__init__(self.msg)


@dataclass
class PhxVertix2D:
    """A 2D vertix."""

    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    x: float
    y: float

    def __eq__(self, other: PhxVertix2D) -> bool:
        TOLERANCE = 0.0001
        return (
            abs(self.x - other.x) < TOLERANCE
            and abs(self.y - other.y) < TOLERANCE
            and self.id_num == other.id_num
        )

    def is_equivalent(self, other: PhxVertix) -> bool:
        TOLERANCE = 0.001
        return abs(self.x - other.x) < TOLERANCE and abs(self.y - other.y) < TOLERANCE

    def __post_init__(self):
        PhxVertix._count += 1
        self.id_num = PhxVertix._count

    @property
    def unique_key(self) -> str:
        """Return a unique key (str) for the Vertex. Used for dicts, welding, etc"""
        return f"{self.x :0.10f}_{self.y :0.10f}"

    def __hash__(self) -> int:
        return hash(self.unique_key)


@dataclass
class PhxVertix:
    _count: ClassVar[int] = 0

    id_num: int = field(init=False, default=0)
    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def __eq__(self, other: PhxVertix) -> bool:
        TOLERANCE = 0.0001
        return (
            abs(self.x - other.x) < TOLERANCE
            and abs(self.y - other.y) < TOLERANCE
            and abs(self.z - other.z) < TOLERANCE
            and self.id_num == other.id_num
        )

    def is_equivalent(self, other: PhxVertix) -> bool:
        TOLERANCE = 0.001
        return (
            abs(self.x - other.x) < TOLERANCE
            and abs(self.y - other.y) < TOLERANCE
            and abs(self.z - other.z) < TOLERANCE
        )

    def __post_init__(self):
        PhxVertix._count += 1
        self.id_num = PhxVertix._count

    @property
    def unique_key(self) -> str:
        """Return a unique key (str) for the Vertex. Used for dicts, welding, etc"""
        return f"{self.x :0.10f}_{self.y :0.10f}_{self.z :0.10f}"

    def __hash__(self) -> int:
        return hash(self.unique_key)


@dataclass
class PhxVector:
    x: float
    y: float
    z: float

    @classmethod
    def from_2_points(cls, _start_pt: PhxVertix, _end_pt: PhxVertix) -> PhxVector:
        """Return a new PhxVector based on a start and end point."""
        return cls(
            _end_pt.x - _start_pt.x, _end_pt.y - _start_pt.y, _end_pt.z - _start_pt.z
        )

    def scale(self, _factor: float) -> None:
        self.x = self.x * _factor
        self.y = self.y * _factor
        self.z = self.z * _factor

    def dot(self, other: PhxVector) -> float:
        """Get the dot product of this vector with another."""
        return self.x * other.x + self.y * other.y + self.z * other.z

    def rotate_around(self, _axis: PhxVector, _angle_deg: float) -> PhxVector:
        """Rotate this vector around an axis by an angle in radians."""
        _cos = math.cos(math.radians(_angle_deg))
        _sin = math.sin(math.radians(_angle_deg))
        _x = (
            (_cos + _axis.x**2 * (1 - _cos)) * self.x
            + (_axis.x * _axis.y * (1 - _cos) - _axis.z * _sin) * self.y
            + (_axis.x * _axis.z * (1 - _cos) + _axis.y * _sin) * self.z
        )
        _y = (
            (_axis.y * _axis.x * (1 - _cos) + _axis.z * _sin) * self.x
            + (_cos + _axis.y**2 * (1 - _cos)) * self.y
            + (_axis.y * _axis.z * (1 - _cos) - _axis.x * _sin) * self.z
        )
        _z = (
            (_axis.z * _axis.x * (1 - _cos) - _axis.y * _sin) * self.x
            + (_axis.z * _axis.y * (1 - _cos) + _axis.x * _sin) * self.y
            + (_cos + _axis.z**2 * (1 - _cos)) * self.z
        )
        return PhxVector(_x, _y, _z)

    def unitize(self) -> None:
        """Convert this vector to a unit vector."""
        magnitude = math.sqrt(self.x**2 + self.y**2 + self.z**2)
        if magnitude > 0:
            self.x /= magnitude
            self.y /= magnitude
            self.z /= magnitude

    def __eq__(self, other: PhxVector) -> bool:
        TOLERANCE = 0.001
        return (
            abs(self.x - other.x) < TOLERANCE
            and abs(self.y - other.y) < TOLERANCE
            and abs(self.z - other.z) < TOLERANCE
        )


@dataclass
class PhxPlane:
    normal_vector: PhxVector
    origin: PhxVertix
    x: PhxVector
    y: PhxVector

    @property
    def normal(self) -> PhxVector:
        return self.normal_vector

    def __eq__(self, other: PhxPlane) -> bool:
        return (
            self.normal_vector == other.normal_vector
            and self.origin == other.origin
            and self.x == other.x
            and self.y == other.y
        )

    def xyz_to_xy(self, point: PhxVertix) -> PhxVertix2D:
        """Get a Point2D in the coordinate system of this plane from a Point3D.

        Copied from Ladybug-Geometry
        https://github.com/ladybug-tools/ladybug-geometry/blob/master/ladybug_geometry/geometry3d/plane.py
        """
        _diff = PhxVector(
            point.x - self.origin.x, point.y - self.origin.y, point.z - self.origin.z
        )
        return PhxVertix2D(self.x.dot(_diff), self.y.dot(_diff))

    def xy_to_xyz(self, point):
        """Get a Point3D from a Point2D in the coordinate system of this plane.

        Cpied from Ladybug-Geometry
        https://github.com/ladybug-tools/ladybug-geometry/blob/master/ladybug_geometry/geometry3d/plane.py
        """
        _u = (self.x.x * point.x, self.x.y * point.x, self.x.z * point.x)
        _v = (self.y.x * point.y, self.y.y * point.y, self.y.z * point.y)
        return PhxVertix(
            self.origin.x + _u[0] + _v[0],
            self.origin.y + _u[1] + _v[1],
            self.origin.z + _u[2] + _v[2],
        )


@dataclass
class PhxLineSegment:
    vertix_1: PhxVertix
    vertix_2: PhxVertix

    @property
    def length(self) -> float:
        return abs(
            math.sqrt(
                (
                    (self.vertix_2.x - self.vertix_1.x) ** 2
                    + (self.vertix_2.y - self.vertix_1.y) ** 2
                    + (self.vertix_2.z - self.vertix_1.z) ** 2
                )
            )
        )

    def __eq__(self, other: PhxLineSegment) -> bool:
        if self.vertix_1 == other.vertix_1 and self.vertix_2 == other.vertix_2:
            return True
        if self.vertix_1 == other.vertix_2 and self.vertix_2 == other.vertix_1:
            return True
        return False


@dataclass
class PhxPolygon:
    """A Polygon surface defined by 3 or more vertices."""

    _count: ClassVar[int] = 0

    _display_name: str
    _area: Optional[float]
    _center: Optional[PhxVertix]
    normal_vector: PhxVector
    plane: PhxPlane

    _vertices: List[PhxVertix] = field(init=False, default_factory=list)
    child_polygon_ids: List[int] = field(init=False, default_factory=list)

    id_num: int = field(init=False, default=0)

    def __eq__(self, other: PhxPolygon) -> bool:
        # -- Check the PolyIds
        if not len(self.child_polygon_ids) == len(other.child_polygon_ids):
            return False
        for _ in self.child_polygon_ids:
            if not _ in other.child_polygon_ids:
                return False

        # # -- Chech the actual vertices
        if not len(self.vertices) == len(other.vertices):
            return False
        # -- For each Vertix, make sure this is an equivalent vertix in the other
        for v in self.vertices:
            if not any((_.is_equivalent(v) for _ in other.vertices)):
                return False

        return True

    @property
    def area(self) -> float:
        if not self._area:
            self._area = self.calculate_area()
        return self._area

    @area.setter
    def area(self, _in: Optional[float]):
        if _in:
            self._area = _in

    @property
    def center(self) -> PhxVertix:
        if not self._center:
            self._center = self.calculate_center()
        return self._center

    @center.setter
    def center(self, _in: Optional[PhxVertix]):
        if _in:
            self._center = _in

    def __post_init__(self):
        PhxPolygon._count += 1
        self.id_num = PhxPolygon._count

    @property
    def display_name(self) -> str:
        if not self._display_name:
            return str(self.id_num)
        else:
            return self._display_name

    @display_name.setter
    def display_name(self, _in: str):
        self._display_name = str(_in)

    @property
    def vertices(self) -> List[PhxVertix]:
        return self._vertices

    @property
    def vertices_id_numbers(self) -> List[int]:
        return [v.id_num for v in self.vertices]

    @property
    def angle_from_horizontal(self) -> float:
        """Return the surface normal's angle (degrees) off horizontal.

        ie: pointing 'up'=0, vertical surface=90, pointing 'down'=180

        https://www.omnicalculator.com/math/angle-between-two-vectors

        Returns:
        -------
            * (float): Degrees. The surface normal angle off horizontal.
        """

        # Get the input Vector's X and Y parts
        a = self.normal_vector
        b = PhxVector(0, 0, 1.0)

        try:
            angle = math.acos(
                (a.x * b.x + a.y * b.y + a.z * b.z)
                / (
                    math.sqrt(a.x**2 + a.y**2 + a.z**2)
                    * math.sqrt(b.x**2 + b.y**2 + b.z**2)
                )
            )
        except ZeroDivisionError:
            angle = 0.0

        return math.degrees(angle)

    @property
    def is_horizontal(self, tolerance: float = 0.0001) -> bool:
        a = self.angle_from_horizontal
        if abs(a) < tolerance:
            return True
        elif abs(180 - abs(a)) < tolerance:
            return True
        return False

    @property
    def is_vertical(self, tolerance: float = 0.0001) -> bool:
        a = self.angle_from_horizontal
        if abs(90 - abs(a)) < tolerance:
            return True
        return False

    @property
    def cardinal_orientation_angle(
        self, _reference_vector: Optional[PhxVector] = None
    ) -> float:
        """Calculate polygon normal's horizontal angle off a reference. By default, the
        reference vector will be (x=0,y=1,z=0) assuming north is y-direction.

        http://frasergreenroyd.com/obtaining-the-angle-between-two-vectors-for-360-degrees/
        Results 0=north, 90=east, 180=south, 270=west

        Arguments:
        ----------
            * _reference_vector: (PhxVector) The reference vector representing
                'North'. By default, will use (x=0, y=1, z=0) as the north vector.

        Returns:
        --------
            (float) Degrees. The clockwise angle off reference of the surface normal.
        """
        if self.is_horizontal:
            return 0.0

        if not _reference_vector:
            _reference_vector = PhxVector(0, 1.0, 0)

        a_x = self.normal_vector.x
        a_y = self.normal_vector.y

        b_x = _reference_vector.x
        b_y = _reference_vector.y

        angle = math.atan2(b_y, b_x) - math.atan2(a_y, a_x)
        angle = math.degrees(angle)

        if angle < 0:
            angle = angle + 360

        return angle

    def add_vertix(self, _phx_vertix: PhxVertix) -> None:
        self.vertices.append(_phx_vertix)

    def add_child_poly_id(self, _child_ids: Union[Collection[int], int]) -> None:
        if not isinstance(_child_ids, Collection):
            _child_ids = (_child_ids,)

        for child_id in _child_ids:
            self.child_polygon_ids.append(child_id)

    def set_vertex(self, _phx_vertix: PhxVertix, index: int):
        """Set a vertex at a specific index."""
        self.vertices[index] = _phx_vertix

    def calculate_area(self) -> float:
        """Calculate the area of the polygon."""
        x_coords = [self.plane.xyz_to_xy(v).x for v in self.vertices]
        y_coords = [self.plane.xyz_to_xy(v).y for v in self.vertices]
        x_coords.append(x_coords[0])
        y_coords.append(y_coords[0])
        return (
            abs(
                sum(
                    [
                        x_coords[i] * y_coords[i + 1] - x_coords[i + 1] * y_coords[i]
                        for i in range(len(x_coords) - 1)
                    ]
                )
            )
            / 2
        )

    def calculate_center(self) -> PhxVertix:
        """Find the center of the polygon."""
        num_vertices = len(self.vertices)
        center = PhxVertix(0, 0, 0)  # Initialize the center point at (0, 0, 0)

        # Sum up the coordinates of all vertices
        for vertex in self.vertices:
            center.x += vertex.x
            center.y += vertex.y
            center.z += vertex.z

        # Divide the sum by the number of vertices to get the average position
        center.x /= num_vertices
        center.y /= num_vertices
        center.z /= num_vertices

        return center

    def scale(self, _scale_factor: float = 1.0) -> None:
        """Scale the polygon by the given factor."""
        for vertex in self.vertices:
            vertex.x -= self.center.x
            vertex.y -= self.center.y
            vertex.z -= self.center.z

            vertex.x *= _scale_factor
            vertex.y *= _scale_factor
            vertex.z *= _scale_factor

            vertex.x += self.center.x
            vertex.y += self.center.y
            vertex.z += self.center.z


@dataclass
class PhxPolygonRectangular(PhxPolygon):
    """A Polygon with additional geometric attributes for rectangular surfaces."""

    vertix_upper_left: Optional[PhxVertix] = field(init=False, default=None)
    vertix_lower_left: Optional[PhxVertix] = field(init=False, default=None)
    vertix_lower_right: Optional[PhxVertix] = field(init=False, default=None)
    vertix_upper_right: Optional[PhxVertix] = field(init=False, default=None)

    VERTEX_ORDER = [
        "vertix_upper_left",
        "vertix_lower_left",
        "vertix_lower_right",
        "vertix_upper_right",
    ]

    def __post_init__(self):
        super().__post_init__()

    @property
    def edge_top(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Top' side of the Polygon (viewed from outside)."""
        if not self.vertix_upper_right or not self.vertix_upper_left:
            raise PolygonEdgeError(self, "top")
        return PhxLineSegment(self.vertix_upper_right, self.vertix_upper_left)

    @property
    def edge_left(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Left' side of the Polygon (viewed from outside)."""
        if not self.vertix_upper_left or not self.vertix_lower_left:
            raise PolygonEdgeError(self, "left")
        return PhxLineSegment(self.vertix_upper_left, self.vertix_lower_left)

    @property
    def edge_bottom(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Bottom' side of the Polygon (viewed from outside)."""
        if not self.vertix_lower_left or not self.vertix_lower_right:
            raise PolygonEdgeError(self, "bottom")
        return PhxLineSegment(self.vertix_lower_left, self.vertix_lower_right)

    @property
    def edge_right(self) -> PhxLineSegment:
        """Returns the PhxLineSegment representing the 'Right' side of the Polygon (viewed from outside)."""
        if not self.vertix_lower_right or not self.vertix_upper_right:
            raise PolygonEdgeError(self, "right")
        return PhxLineSegment(self.vertix_lower_right, self.vertix_upper_right)

    @property
    def width(self) -> float:
        return self.edge_top.length

    @property
    def height(self) -> float:
        return self.edge_left.length

    def add_vertix(self, _phx_vertix: PhxVertix) -> None:
        print(
            f'Method "add_vertix()" is not allowed for {self.__class__.__name__} objects.'
            f"Please assign the vertix corners directly (vertix_upper_left-left, vertix_lower_left-right, etc..)"
        )

    @property
    def vertices(self) -> List[PhxVertix]:
        """Return a List of the PhxPolygonRectangle Vertices (counter-clockwise from upper-left)."""
        corner_vertices = [getattr(self, v) for v in self.VERTEX_ORDER]
        # filter out any 'None' vertices
        return [v for v in corner_vertices if v]

    def set_vertex(self, _phx_vertix: PhxVertix, index: int):
        """Set a vertex at a specific index."""
        vert_name = self.VERTEX_ORDER[index]
        setattr(self, vert_name, _phx_vertix)

    @property
    def area(self) -> float:
        """Returns the area of the rectangular surface."""
        return self.width * self.height


@dataclass
class PhxGraphics3D:
    polygons: List[PhxPolygon] = field(default_factory=list)

    @property
    def vertices(self) -> List[PhxVertix]:
        """Returns a sorted list with all of the unique vertix objects of all the polygons in the collection."""
        return sorted(
            {vertix for polygon in self.polygons for vertix in polygon.vertices},
            key=lambda _: _.id_num,
        )

    def add_polygons(self, _polygons: Union[Collection[PhxPolygon], PhxPolygon]) -> None:
        """Adds a new Polygon object to the collection"""

        if not isinstance(_polygons, Collection):
            _polygons = [_polygons]

        for polygon in _polygons:
            self.polygons.append(polygon)

    def get_polygons_by_id(self, _ids: Collection[int]) -> List[PhxPolygon]:
        """Returns a sorted list of polygons in the collection matching the IDs supplied.

        Arguments:
        ----------
            * _ids: (Collection[int]): A collection of one or more id_nums to look for.

        Returns:
        --------
            * List[PhxPolygon]: A sorted (by display_name) list of all the
                polygons with matching id_nums.
        """

        if not isinstance(_ids, Collection):
            _ids = {_ids}

        return sorted(
            [p for p in self.polygons if p.id_num in _ids], key=lambda _: _.display_name
        )

    def __bool__(self) -> bool:
        return bool(self.polygons)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP 'Overview' Worksheet."""

from __future__ import annotations

from ph_units.unit_type import Unit

from PHX.xl.xl_app import XLConnection
from PHX.PHPP.phpp_localization import shape_model as shp
from PHX.PHPP.sheet_io.io_exceptions import PHPPDataMissingException


class OverviewBasicData:
    def __init__(self, _host, _xl: XLConnection, _shape: shp.OverviewBasicData):
        self.host = _host
        self.xl = _xl
        self.shape = _shape

    def get_num_dwellings(self) -> int:
        """Return the Total Net Interior Volume (Vn50)

        They put res and non-res dwelling num on different columns?
        """
        address_res = self.shape.address_number_dwellings_res
        val_res = self.xl.get_single_data_item(self.host.worksheet_name, address_res)
        if val_res:
            return int(val_res)

        address_nonres = self.shape.address_number_dwellings_nonres
        val_nonres = self.xl.get_single_data_item(
            self.host.worksheet_name, address_nonres
        )
        if val_nonres:
            return int(val_nonres)

        raise PHPPDataMissingException(
            self.host.worksheet_name, [address_res, address_nonres]
        )

    def get_num_occupants(self) -> float:
        """Return the number of occupants.

        Note: They put res and non-res occupancy on different columns?
        """
        address_res = self.shape.address_number_occupants_res
        val_res = self.xl.get_single_data_item(self.host.worksheet_name, address_res)
        if val_res:
            return float(val_res)

        address_nonres = self.shape.address_number_occupants_nonres
        val_nonres = self.xl.get_single_data_item(
            self.host.worksheet_name, address_nonres
        )
        if val_nonres:
            return float(val_nonres)

        raise PHPPDataMissingException(
            self.host.worksheet_name, [address_res, address_nonres]
        )

    def get_project_name(self) -> str:
        """Return the name of the Project / Building"""
        address = self.shape.address_project_name
        val = self.xl.get_single_data_item(self.host.worksheet_name, address)
        if val:
            return str(val)

        raise PHPPDataMissingException(self.host.worksheet_name, address)


class OverviewVentilation:
    def __init__(self, _host, _xl: XLConnection, _shape: shp.OverviewVentilation) -> None:
        self.host = _host
        self.xl = _xl
        self.shape = _shape

    def get_vn50(self) -> Unit:
        """Return the Total Net Interior Volume (Vn50)"""
        address = self.shape.vn50.xl_range
        val = self.xl.get_single_data_item(self.host.worksheet_name, address)
        return Unit(float(val or 0.0), str(self.shape.vn50.unit))


class OverviewBuildingEnvelope:
    """IO Controller for the PHPP 'Overview' worksheet."""

    def __init__(
        self, _host, _xl: XLConnection, _shape: shp.OverviewBuildingEnvelope
    ) -> None:
        self.host = _host
        self.xl = _xl
        self.shape = _shape

    def get_area_envelope(self) -> Unit:
        """Return the Total Envelope Area [M2]"""
        address = self.shape.address_area_envelope.xl_range
        val = self.xl.get_single_data_item(self.host.worksheet_name, address)
        return Unit(float(val or 0.0), str(self.shape.address_area_envelope.unit))

    def get_area_tfa(self) -> Unit:
        """Return the Total TFA [M2]"""
        address = self.shape.address_area_tfa.xl_range
        val = self.xl.get_single_data_item(self.host.worksheet_name, address)
        return Unit(float(val or 0.0), str(self.shape.address_area_tfa.unit))


class Overview:
    """IO Controller for the PHPP 'Overview' worksheet."""

    def __init__(self, _xl: XLConnection, _shape: shp.Overview) -> None:
        self.xl = _xl
        self.shape = _shape
        self.basic_data = OverviewBasicData(self, _xl, self.shape.basic_data)
        self.building_envelope = OverviewBuildingEnvelope(
            self, _xl, self.shape.building_envelope
        )
        self.ventilation = OverviewVentilation(self, _xl, self.shape.ventilation)

    @property
    def worksheet_name(self) -> str:
        return self.shape.name

    def get_area_envelope(self) -> Unit:
        """Return the Total Envelope Area [M2 | FT2]"""
        return self.building_envelope.get_area_envelope()

    def get_area_tfa(self) -> Unit:
        """Return the Total TFA [M2 | FT2]"""
        return self.building_envelope.get_area_tfa()

    def get_net_interior_volume(self) -> Unit:
        """Return the Total Net Interior Volume (Vn50) [M3 | FT3]"""
        return self.ventilation.get_vn50()

    def get_number_of_dwellings(self) -> int:
        """Return the total number of dwellings."""
        return self.basic_data.get_num_dwellings()

    def get_number_of_occupants(self) -> float:
        """Return the total number of occupants."""
        return self.basic_data.get_num_occupants()

    def get_project_name(self) -> str:
        """Return the Name of the Project."""
        return self.basic_data.get_project_name()



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to create PHX-HVAC objects from Honeybee-Energy-PH HVAC"""

from typing import TypeVar, Union

from honeybee_energy_ph.hvac import (
    ventilation,
    heating,
    cooling,
    _base,
    ducting,
    supportive_device,
)
from honeybee_energy_ph.hvac.renewable_devices import (
    PhRenewableEnergyDevice,
    PhPhotovoltaicDevice,
)
from PHX.model.enums.hvac import PhxVentDuctType, PhxSupportiveDeviceType
from PHX.model import hvac
from PHX.model.hvac.heating import AnyPhxHeater
from PHX.model.hvac.ventilation import AnyPhxVentilation
from PHX.model.hvac.cooling import AnyPhxCooling
from PHX.model.hvac.renewable_devices import PhxDevicePhotovoltaic, AnyRenewableDevice

T = TypeVar("T", bound=Union[AnyPhxVentilation, AnyPhxHeater, AnyPhxCooling])


def _transfer_attributes(_hbeph_obj: _base._PhHVACBase, _phx_obj: T) -> T:
    """Copy the common attributes from a Honeybee-Energy-PH obj to a PHX-object

    Arguments:
    ---------
        * _hbeph_obj (_base._PhHVACBase): A Honeybee-Energy-PH HVAC Object.
        * _phx_obj (mech_equip.PhxMechanicalEquipment): A PHX-Mechanical
            Equipment object.

    Returns:
    --------
        * (mech_equip.PhxMechanicalEquipment): The input PHX Mechanical Equipment
            object, with its attribute values set to match the Honeybee-PH object.
    """

    # -- Copy the base scope attributes from HB->PHX
    hb_attrs = {
        attr_name for attr_name in dir(_hbeph_obj) if not attr_name.startswith("_")
    }
    phx_base_attrs = {
        attr_name for attr_name in dir(_phx_obj) if not attr_name.startswith("_")
    }

    for attr_nm in hb_attrs.intersection(phx_base_attrs):
        setattr(_phx_obj, attr_nm, getattr(_hbeph_obj, attr_nm))

    # -- Also copy attrs to PHX.params for objects which have them
    if _phx_obj.params:
        phx_params_attrs = {
            attr_name
            for attr_name in dir(_phx_obj.params)
            if not attr_name.startswith("_")
        }
        for attr_nm in hb_attrs.intersection(phx_params_attrs):
            setattr(_phx_obj.params, attr_nm, getattr(_hbeph_obj, attr_nm))

    return _phx_obj


# -----------------------------------------------------------------------------
# -- Ventilation


def _default_ventilator_name(_hr: float, _mr: float) -> str:
    """Return a Ventilation Unit name

    Arguments:
    ----------
        * _hr (float): The Heat Recovery efficiency of the unit.
        * _mr (float): The Moisture Recovery efficiency of the unit.
    Returns:
    --------
        * (str)
    """
    return f"Ventilator: {_hr*100 :0.0f}%-HR, {_mr*100 :0.0f}%-MR"


def build_phx_ventilator(
    _hbeph_vent_sys: ventilation.PhVentilationSystem,
) -> hvac.PhxDeviceVentilator:
    """Returns a new Fresh-Air Ventilator built from the hb-energy hvac parameters.

    This will look at the Space's Host-Room .properties.energy.hvac for data.

    Arguments:
    ----------
        *_hbeph_vent (ventilation.PhVentilationSystem): The Honeybee-PH Ventilation System
            to build the PHX-Ventilation from.

    Returns:
    --------
        * (mech_equip.Ventilator): The new Passive House Ventilator created.
    """

    phx_vent = hvac.PhxDeviceVentilator()
    _id_num = phx_vent.id_num  # preserver so 'transfer_attributes' doesn't kill it...

    if not _hbeph_vent_sys.ventilation_unit:
        return phx_vent
    phx_vent = _transfer_attributes(_hbeph_vent_sys, phx_vent)
    phx_vent = _transfer_attributes(_hbeph_vent_sys.ventilation_unit, phx_vent)

    # -- Sort out the Display name to use
    if not _hbeph_vent_sys.ventilation_unit:
        ventilator_name = _default_ventilator_name(0.0, 0.0)
    elif _hbeph_vent_sys.ventilation_unit.display_name == "_unnamed_ventilator_":
        ventilator_name = _default_ventilator_name(
            _hbeph_vent_sys.ventilation_unit.sensible_heat_recovery,
            _hbeph_vent_sys.ventilation_unit.latent_heat_recovery,
        )
    else:
        ventilator_name = _hbeph_vent_sys.ventilation_unit.display_name

    phx_vent.display_name = ventilator_name
    phx_vent.id_num = _id_num

    return phx_vent


def build_phx_duct(
    _hbph_duct: ducting.PhDuctElement, _vent_unit_id: int
) -> hvac.PhxDuctElement:
    """Return a new PHX Ventilation Duct based on an HB-PH Duct.

    Arguments:
    ----------
        * _hbph_duct (ducting.PhDuctElement): The HB-PH DuctElement to use as
            the source.
        * _vent_unit_id (int): The ID-Number of the ventilation unit this duct
            is assigned to.

    Returns:
    --------
        * (hvac.PhxDuctElement)
    """

    phx_duct = hvac.PhxDuctElement(
        _hbph_duct.identifier,
        _hbph_duct.display_name,
        _vent_unit_id,
    )

    phx_duct.duct_type = PhxVentDuctType(
        _hbph_duct.duct_type,
    )

    for hbph_duct_segment in _hbph_duct.segments:
        phx_duct.add_segment(
            hvac.PhxDuctSegment(
                hbph_duct_segment.identifier,
                hbph_duct_segment.display_name,
                hbph_duct_segment.geometry,
                hbph_duct_segment.diameter,
                hbph_duct_segment.height,
                hbph_duct_segment.width,
                hbph_duct_segment.insulation_thickness,
                hbph_duct_segment.insulation_conductivity,
                hbph_duct_segment.insulation_reflective,
            )
        )

    return phx_duct


def build_phx_exh_vent_dryer(
    _hbeph_exhaust_vent: ventilation.ExhaustVentDryer,
) -> hvac.PhxExhaustVentilatorDryer:
    """Create a new PHX Exhaust Vent (Dryer) based on the attributes of an HBPH-Vent Dryer

    Arguments:
    ----------
        * _hbeph_exhaust_vent (ventilation.ExhaustVentDryer): The HBPH Exhaust Vent. Device.

    Returns:
    --------
        * (hvac.PhxExhaustVentilatorDryer): The new PHX Exhaust Vent Device (Dryer)
    """

    obj = hvac.PhxExhaustVentilatorDryer()
    obj.display_name = _hbeph_exhaust_vent.display_name
    obj.params.exhaust_flow_rate_m3h = _hbeph_exhaust_vent.exhaust_flow_rate_m3s * 60 * 60
    return obj


def build_phx_exh_vent_kitchen_vent(
    _hbeph_exhaust_vent: ventilation.ExhaustVentKitchenHood,
) -> hvac.PhxExhaustVentilatorRangeHood:
    """Create a new PHX Exhaust Vent (Kitchen Hood) based on the attributes of an HBPH-Vent Kitchen Hood

    Arguments:
    ----------
        * _hbeph_exhaust_vent (ventilation.ExhaustVentKitchenHood): The HBPH Exhaust Vent. Device.

    Returns:
    --------
        * (hvac.PhxExhaustVentilatorRangeHood): The new PHX Exhaust Vent Device (Dryer)
    """

    obj = hvac.PhxExhaustVentilatorRangeHood()
    obj.display_name = _hbeph_exhaust_vent.display_name
    obj.params.exhaust_flow_rate_m3h = _hbeph_exhaust_vent.exhaust_flow_rate_m3s * 60 * 60
    return obj


def build_phx_exh_vent_use_defined(
    _hbeph_exhaust_vent: ventilation.ExhaustVentUserDefined,
) -> hvac.PhxExhaustVentilatorUserDefined:
    """Create a new PHX Exhaust Vent (User Defined) based on the attributes of an HBPH-Vent User Defined

    Arguments:
    ----------
        * _hbeph_exhaust_vent (ventilation.ExhaustVentUserDefined): The HBPH Exhaust Vent. Device.

    Returns:
    --------
        * (hvac.PhxExhaustVentilatorUserDefined): The new PHX Exhaust Vent Device (Dryer)
    """

    obj = hvac.PhxExhaustVentilatorUserDefined()
    obj.display_name = _hbeph_exhaust_vent.display_name
    obj.params.exhaust_flow_rate_m3h = _hbeph_exhaust_vent.exhaust_flow_rate_m3s * 60 * 60
    obj.params.annual_runtime_minutes = _hbeph_exhaust_vent.annual_runtime_minutes
    return obj


def build_phx_exhaust_vent_device(
    _hbeph_exhaust_vent: ventilation._ExhaustVentilatorBase,
) -> hvac.AnyPhxExhaustVent:
    """Build a new PHX Exhaust Ventilation Mechanical Device.

    Arguments:
    ----------
        *_hbeph_exhaust_vent (ventilation._ExhaustVentilatorBase): The Honeybee-PH
            Exhaust Ventilation Device to build the PHX-Ventilation Device from.

    Returns:
    --------
        * (mech.AnyPhxExhaustVent): A new PHX Exhaust ventilation device.
    """

    # -- Mapping Honeybee-PH -> PHX types
    phx_device_classes = {
        "ExhaustVentDryer": build_phx_exh_vent_dryer,
        "ExhaustVentKitchenHood": build_phx_exh_vent_kitchen_vent,
        "ExhaustVentUserDefined": build_phx_exh_vent_use_defined,
    }

    # -- Get and build the right device type
    phx_device = phx_device_classes[_hbeph_exhaust_vent.device_class_name](
        _hbeph_exhaust_vent
    )
    return phx_device


# -----------------------------------------------------------------------------
# -- Heating


def build_phx_heating_electric(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterElectric:
    phx_obj = hvac.PhxHeaterElectric()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_fossil_boiler(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterBoilerFossil:
    phx_obj = hvac.PhxHeaterBoilerFossil()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_wood_boiler(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterBoilerWood:
    phx_obj = hvac.PhxHeaterBoilerWood()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_district(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterDistrictHeat:
    phx_obj = hvac.PhxHeaterDistrictHeat()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_annual(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterHeatPumpAnnual:
    phx_obj = hvac.PhxHeaterHeatPumpAnnual()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_monthly(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterHeatPumpMonthly:
    phx_obj = hvac.PhxHeaterHeatPumpMonthly()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_hp_combined(
    _hbeph_heater: heating.PhHeatingSystem,
) -> hvac.PhxHeaterHeatPumpCombined:
    phx_obj = hvac.PhxHeaterHeatPumpCombined()
    phx_obj = _transfer_attributes(_hbeph_heater, phx_obj)
    phx_obj.usage_profile.space_heating = True
    return phx_obj


def build_phx_heating_device(_htg_sys: heating.PhHeatingSystem) -> hvac.PhxHeatingDevice:
    """Returns a new PHX-Heating-Device based on an input Honeybee-PH Heating System.

    Arguments:
    ----------
        * _htg_sys (heating.PhHeatingSystem): The Honeybee-PH Heating System to build
            the new PHX Heating system from.

    Returns:
    --------
        * (mech.heating.PhxHeatingDevice): The new PHX Heating System created.
    """

    # -- Mapping Honeybee-PH -> PHX types
    phx_heater_classes = {
        "PhHeatingDirectElectric": build_phx_heating_electric,
        "PhHeatingFossilBoiler": build_phx_heating_fossil_boiler,
        "PhHeatingWoodBoiler": build_phx_heating_wood_boiler,
        "PhHeatingDistrict": build_phx_heating_district,
        "PhHeatingHeatPumpAnnual": build_phx_heating_hp_annual,
        "PhHeatingHeatPumpRatedMonthly": build_phx_heating_hp_monthly,
        "PhHeatingHeatPumpCombined": build_phx_heating_hp_combined,
    }

    # -- Get and build the right heater equipment type
    phx_heater = phx_heater_classes[_htg_sys.heating_type](_htg_sys)
    return phx_heater


def build_phx_heating_sys(_htg_sys: heating.PhHeatingSystem) -> hvac.AnyPhxHeater:
    """

    Arguments:
    ----------
        * _htg_sys (heating.PhHeatingSystem): The Honeybee-PH Heating System to build
            the new PHX Heating system from.

    Returns:
    --------
        * (mech.PhxHeatingDevice):
    """

    phx_htg_device = build_phx_heating_device(_htg_sys)

    # TODO: Distribution...

    return phx_htg_device


# -----------------------------------------------------------------------------
# --- Cooling


def build_phx_cooling_ventilation(
    _hbeph_cooling: cooling.PhCoolingSystem,
) -> hvac.PhxCoolingVentilation:
    phx_obj = hvac.PhxCoolingVentilation()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_recirculation(
    _hbeph_cooling: cooling.PhCoolingSystem,
) -> hvac.PhxCoolingRecirculation:
    phx_obj = hvac.PhxCoolingRecirculation()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_dehumidification(
    _hbeph_cooling: cooling.PhCoolingSystem,
) -> hvac.PhxCoolingDehumidification:
    phx_obj = hvac.PhxCoolingDehumidification()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_panel(
    _hbeph_cooling: cooling.PhCoolingSystem,
) -> hvac.PhxCoolingPanel:
    phx_obj = hvac.PhxCoolingPanel()
    phx_obj = _transfer_attributes(_hbeph_cooling, phx_obj)
    phx_obj.usage_profile.cooling = True
    return phx_obj


def build_phx_cooling_device(_clg_sys: cooling.PhCoolingSystem) -> hvac.PhxCoolingDevice:
    """Returns a new PHX-Cooling-Device based on an input Honeybee-PH cooling System.

    Arguments:
    ----------
        * _clg_sys (cooling.PhCoolingSystem): The Honeybee-PH Cooling-System to build
            the new PHX-Cooling-System from.

    Returns:
    --------
        * (mech.cooling.PhxHeatingDevice): The new PHX cooling System created.
    """

    # -- Mapping Honeybee-PH -> PHX types
    phx_cooling_classes = {
        "PhCoolingVentilation": build_phx_cooling_ventilation,
        "PhCoolingRecirculation": build_phx_cooling_recirculation,
        "PhCoolingDehumidification": build_phx_cooling_dehumidification,
        "PhCoolingPanel": build_phx_cooling_panel,
    }

    # -- Get and build the right heater equipment type
    phx_cooling = phx_cooling_classes[_clg_sys.cooling_class_name](_clg_sys)
    return phx_cooling


def build_phx_cooling_sys(_htg_sys: cooling.PhCoolingSystem) -> hvac.AnyPhxCooling:
    """

    Arguments:
    ----------
        * _htg_sys (cooling.PhCoolingSystem): The Honeybee-PH cooling System to build
            the new PHX cooling system from.

    Returns:
    --------
        * (mech.PhxCoolingDevice):
    """

    phx_htg_device = build_phx_cooling_device(_htg_sys)

    # TODO: Distribution...

    return phx_htg_device


# -----------------------------------------------------------------------------
# --- Supportive Devices


def build_phx_supportive_device(
    _hbeph_supportive_device: supportive_device.PhSupportiveDevice,
) -> hvac.PhxSupportiveDevice:
    """Build a new PhxSupportiveDevice Device.

    Arguments:
    ----------
        *_hbeph_supportive_device (supportive_device.PhSupportiveDevice): The Honeybee-PH
            supportive device to build the PHX-Ventilation Device from.

    Returns:
    --------
        * (hvac.PhxSupportiveDevice): A new PHX Supportive Device.
    """
    phx_device = hvac.PhxSupportiveDevice()

    # -- basics
    phx_device.identifier = _hbeph_supportive_device.identifier
    phx_device.display_name = _hbeph_supportive_device.display_name
    phx_device.device_type = PhxSupportiveDeviceType(_hbeph_supportive_device.device_type)
    phx_device.quantity = _hbeph_supportive_device.quantity

    # -- params
    phx_device.params.in_conditioned_space = _hbeph_supportive_device.in_conditioned_space
    phx_device.params.norm_energy_demand_W = _hbeph_supportive_device.norm_energy_demand_W
    phx_device.params.annual_period_operation_khrs = (
        _hbeph_supportive_device.annual_period_operation_khrs
    )

    return phx_device


# -----------------------------------------------------------------------------
# -- Renewable Devices


def build_phx_photovoltaic_device(_hbph_d: PhPhotovoltaicDevice) -> PhxDevicePhotovoltaic:
    """Build a new PhxPhotovoltaicDevice Device from a Honeybee-PH Photovoltaic Device."""
    d = PhxDevicePhotovoltaic()
    d.display_name = _hbph_d.display_name
    d.params.photovoltaic_renewable_energy = _hbph_d.photovoltaic_renewable_energy
    d.params.onsite_utilization_factor = _hbph_d.utilization_factor
    d.params.array_size = _hbph_d.array_size
    return d


def build_phx_renewable_device(_device: PhRenewableEnergyDevice) -> AnyRenewableDevice:
    """Build a new PHX-Renewable-Device Device from a Honeybee-PH Renewable-Device."""
    # -- Mapping Honeybee-PH -> PHX types
    phx_renewable_classes = {
        "PhPhotovoltaicDevice": build_phx_photovoltaic_device,
    }

    # -- Get and build the right heater equipment type
    phx_device = phx_renewable_classes[_device.device_typename](_device)
    return phx_device



from typing import Any, Dict
from PHX.model import project
from PHX.from_WUFI_XML.phx_schemas import _PhxProject
from PHX.from_WUFI_XML.wufi_file_schema import WUFIplusProject


def convert_WUFI_XML_to_PHX_project(_model: WUFIplusProject) -> project.PhxProject:
    phx_project = _PhxProject(_model)

    return phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for finding / loading the right PHPP-ShapeFile."""


import pathlib
import os

from PHX.PHPP.phpp_localization.shape_model import PhppShape
from PHX.PHPP.phpp_model import version
from PHX.xl import xl_app


def phpp_version_as_file_name(version: version.PHPPVersion):
    """Return the version as a file-name."""
    return f"{version.language}_{version.number_major}_{version.number_minor}"


def get_shape_filepath(version: version.PHPPVersion, _shape_file_directory: pathlib.Path):
    """Returns the path to the PHPP Shape File based on the language and PHPP version found.

    Arguments:
    ----------
        * _xl (xl_app.XLConnection): The XL connection to use to find the version data.
        * _shape_file_directory ():

    Returns:
    --------
        * (pathlib.Path): The path to the correct PHPP Shape File.
    """
    shape_file_name = phpp_version_as_file_name(version)
    shape_file_name = f"{shape_file_name}.json"
    shape_file_path = pathlib.Path(_shape_file_directory, shape_file_name)

    if not os.path.exists(shape_file_path):
        raise FileNotFoundError(
            f'\n\tError: The PHPP shapefile "{shape_file_path}" was not found?'
        )

    return shape_file_path


def get_phpp_shape(_xl: xl_app.XLConnection, version: version.PHPPVersion) -> PhppShape:
    """Return the PhppShape Object."""

    shape_file_dir = pathlib.Path(__file__).parent
    phpp_shape_filepath = get_shape_filepath(version, shape_file_dir)
    _xl.output(f"Loading PHPP Shapefile: {phpp_shape_filepath}")
    phpp_shape = PhppShape.parse_file(phpp_shape_filepath)

    return phpp_shape



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Valid 'types' for PHIUS Certification Settings."""

from enum import Enum


class PhiusCertificationBuildingCategoryType(Enum):
    RESIDENTIAL_BUILDING = 1
    NONRESIDENTIAL_BUILDING = 2


class PhiusCertificationBuildingUseType(Enum):
    RESIDENTIAL = 1
    OFFICE = 4
    SCHOOL = 5
    OTHER = 6
    UNDEFINED = 7


class PhiusCertificationBuildingStatus(Enum):
    IN_PLANNING = 1
    UNDER_CONSTRUCTION = 2
    COMPLETE = 3


class PhiusCertificationBuildingType(Enum):
    NEW_CONSTRUCTION = 1
    RETROFIT = 2
    MIXED = 3

class PhiusCertificationProgram(Enum):
    DEFAULT = 1
    PHIUS_2015 = 2
    PHIUS_2018 = 3
    ITALIAN  = 4
    PHIUS_2018_CORE = 5
    PHIUS_2018_ZERO = 6
    PHIUS_2021_CORE = 7
    PHIUS_2021_ZERO = 8


# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Class for managing PHPP Version data."""


class PHPPVersion:
    """Manage the PHPP Version number and language information."""

    def __init__(self, _number_major: str, _number_minor: str, _language: str):
        self.number_major = self.clean_input(_number_major)
        self.number_minor = self.clean_input(_number_minor)
        self.language = self.clean_input(_language)

    def clean_input(self, _input):
        """Upper, strip, replace spaces."""
        return str(_input).upper().strip().replace(" ", "").replace(".", "_")

    def number(self) -> str:
        """Return the full version number (ie: "9.6", "10.4", etc..)"""
        return f"{self.number_major}.{self.number_minor}"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Program Class for organizing Fresh-Air Ventilation Data."""

from __future__ import annotations
from dataclasses import dataclass, field

from PHX.model.loads import ventilation as vent_loads
from PHX.model.schedules import ventilation as vent_schedules


@dataclass
class PhxProgramVentilation:
    """A PHX Program for the Fresh-Air Ventilation with a load and schedule."""

    display_name: str = "Unnamed_Ventilation_Program"
    load: vent_loads.PhxLoadVentilation = field(
        default_factory=vent_loads.PhxLoadVentilation
    )
    schedule: vent_schedules.PhxScheduleVentilation = field(
        default_factory=vent_schedules.PhxScheduleVentilation
    )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""The data from an existing PHPP Non-Res Lighting Entry row."""

from dataclasses import dataclass
import re
from typing import List, Optional, Dict

from PHX.xl import xl_data
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class ExistingLightingRow:
    shape: shape_model.ElecNonRes
    data: list

    @property
    def room_name(self) -> str:
        col_letter = str(self.shape.lighting_rows.inputs.room_zone_name)
        col_number_as_index = xl_data.xl_ord(col_letter) - 65
        return self.data[col_number_as_index]

    @property
    def utilization_profile_phpp_id(self) -> str:
        col_letter = str(self.shape.lighting_rows.inputs.utilization_profile)
        col_number_as_index = xl_data.xl_ord(col_letter) - 65
        return self.data[col_number_as_index]

    @property
    def utilization_profile_number(self) -> str:
        phpp_id = self.utilization_profile_phpp_id
        pattern = r"(\d+)-(.*)"

        match = re.search(pattern, phpp_id)
        if match:
            return match.group(1)
        else:
            msg = f"Error: Unexpected utilization profile name format? '{phpp_id}'"
            raise Exception(msg)

    @property
    def utilization_profile_name(self) -> str:
        phpp_id = self.utilization_profile_phpp_id
        pattern = r"(\d+)-(.*)"

        match = re.search(pattern, phpp_id)
        if match:
            return str(match.group(2))
        else:
            msg = f"Error: Unexpected utilization profile name format? '{phpp_id}'"
            raise Exception(msg)

    @property
    def installed_power(self) -> str:
        col_letter = str(self.shape.lighting_rows.inputs.installed_power)
        col_number_as_index = xl_data.xl_ord(col_letter) - 65
        return self.data[col_number_as_index]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Building Classes"""

from __future__ import annotations
from typing import ClassVar, List, Sequence, Union, Set, Dict, ValuesView, Any, Optional
from dataclasses import dataclass, field
from collections import defaultdict
from functools import reduce
import operator

from PHX.model import elec_equip, geometry, spaces
from PHX.model.components import (
    PhxComponentAperture,
    PhxComponentOpaque,
    PhxComponentThermalBridge,
)
from PHX.model.hvac import collection
from PHX.model.programs import occupancy
from PHX.model.enums.building import SpecificHeatCapacity


@dataclass
class PhxZone:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    display_name: str = ""
    volume_gross: float = 0.0
    volume_net: float = 0.0
    weighted_net_floor_area: float = 0.0
    clearance_height: float = 2.5
    res_occupant_quantity: float = 0.0
    res_number_bedrooms: int = 0
    res_number_dwellings: int = 0
    specific_heat_capacity: SpecificHeatCapacity = SpecificHeatCapacity.LIGHTWEIGHT

    spaces: List[spaces.PhxSpace] = field(default_factory=list)
    _thermal_bridges: Dict[str, PhxComponentThermalBridge] = field(default_factory=dict)

    # TODO: see if this can be safely removed?
    # ---------------------------------------------------------
    occupancy: Optional[occupancy.PhxProgramOccupancy] = None
    # ---------------------------------------------------------

    lighting: Any = None
    elec_equipment_collection: elec_equip.PhxElectricDeviceCollection = field(
        default_factory=elec_equip.PhxElectricDeviceCollection
    )
    exhaust_ventilator_collection: collection.PhxExhaustVentilatorCollection = field(
        default_factory=collection.PhxExhaustVentilatorCollection
    )

    def add_thermal_bridge(self, _thermal_bridge: PhxComponentThermalBridge) -> None:
        """Add a new PhxComponentThermalBridge to the PhxZone."""
        self._thermal_bridges[_thermal_bridge.identifier] = _thermal_bridge

    def add_thermal_bridges(
        self,
        _thermal_bridges: Union[
            PhxComponentThermalBridge, Sequence[PhxComponentThermalBridge]
        ],
    ) -> None:
        """Add a new PhxComponentThermalBridge (or list of Bridges) to the PhxZone."""
        if not isinstance(_thermal_bridges, Sequence):
            _thermal_bridges = (_thermal_bridges,)

        for tb in _thermal_bridges:
            self._thermal_bridges[tb.identifier] = tb

        return None

    @property
    def thermal_bridges(self) -> ValuesView[PhxComponentThermalBridge]:
        """Return all of the PhxComponentThermalBridge objects in the PhxZone."""
        return self._thermal_bridges.values()

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count


@dataclass
class PhxBuilding:
    """PHX Building Class"""

    # -- Only opaque components (and shades) are stored in the _components list
    # -- as the apertures are stored in the opaque components themselves
    _components: List[PhxComponentOpaque] = field(default_factory=list)
    zones: List[PhxZone] = field(default_factory=list)

    @property
    def weighted_net_floor_area(self) -> float:
        """Returns the total weighted net floor area of all zones in the PhxBuilding."""
        return sum(z.weighted_net_floor_area for z in self.zones)

    @property
    def net_volume(self) -> float:
        """Returns the total net-volume of all the zones in the PhxBuilding."""
        return sum(z.volume_net for z in self.zones)

    def add_components(
        self, _components: Union[PhxComponentOpaque, Sequence[PhxComponentOpaque]]
    ) -> None:
        """Add new PHX Components to the PhxBuilding."""
        if not isinstance(_components, Sequence):
            _components = (_components,)

        for compo in _components:
            self._components.append(compo)

    def add_component(self, _component: PhxComponentOpaque) -> None:
        """Add a new PHX Components to the PhxBuilding."""
        self._components.append(_component)

    def add_zones(self, _zones: Union[PhxZone, Sequence[PhxZone]]) -> None:
        """Add a new PhxZone to the PhxBuilding."""
        if not isinstance(_zones, Sequence):
            _zones = (_zones,)

        for zone in _zones:
            self.zones.append(zone)

    def add_zone(self, _zone: PhxZone) -> None:
        """Add a new PhxZone to the PhxBuilding."""
        self.zones.append(_zone)

    def merge_opaque_components_by_assembly(self) -> None:
        """Merge together all the Opaque-Components in the Building if they gave the same Attributes."""
        # -- Group the opaque components by their unique key / type
        new_component_groups = defaultdict(list)
        for c in self.opaque_components:
            new_component_groups[c.unique_key].append(c)

        # -- Create new components from the group
        grouped_opaque_components: List[PhxComponentOpaque] = []
        for component_group in new_component_groups.values():
            grouped_opaque_components.append(reduce(operator.add, component_group))

        # -- Reset the Building's Components
        self._components = grouped_opaque_components

    def merge_aperture_components_by_assembly(self) -> None:
        """Merge together all the Aperture-Components in the Building if they gave the same Attributes."""
        # -- Group the aperture components by their unique key / type
        new_components: List[PhxComponentOpaque] = []
        for c in self.opaque_components:
            new_component_groups = defaultdict(list)
            for a in c.apertures:
                new_component_groups[a.unique_key].append(a)

            # -- Create new components from the groups
            grouped_aperture_components = []
            for component_group in new_component_groups.values():
                grouped_aperture_components.append(reduce(operator.add, component_group))

            # -- Reset the Components's Apertures
            c.apertures = grouped_aperture_components
            new_components.append(c)

        # -- Reset the Building's Components
        self._components = new_components

    @property
    def all_components(self) -> List[Union[PhxComponentOpaque, PhxComponentAperture]]:
        """Return a list of all the Opaque and Aperture Components in the Building.

        Returns:
        --------
            * (List[Union[PhxComponentOpaque, PhxComponentAperture]]) A list of all
                the opaque and aperture components.
        """
        all_components: List[Union[PhxComponentOpaque, PhxComponentAperture]] = []
        for c in self._components:
            all_components += c.apertures
            all_components.append(c)
        return sorted(all_components, key=lambda c: c.id_num)

    @property
    def aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the aperture components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the aperture components.
        """
        return [ap for c in self.opaque_components for ap in c.apertures]

    @property
    def wall_aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the wall aperture (window) components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the wall aperture components.
        """
        return sorted(
            [ap for c in self.above_grade_wall_components for ap in c.apertures],
            key=lambda ap: ap.display_name,
        )

    @property
    def roof_aperture_components(self) -> List[PhxComponentAperture]:
        """Returns a sorted list (by display name) of all the roof aperture (skylight) components in the building.

        Returns:
        --------
            * (List[PhxComponentAperture]) A sorted list of all the roof aperture components.
        """
        return sorted(
            [ap for c in self.roof_components for ap in c.apertures],
            key=lambda ap: ap.display_name,
        )

    @property
    def opaque_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the opaque non-shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the opaque components.
        """
        return sorted(
            [c for c in self._components if not c.is_shade], key=lambda _: _.display_name
        )

    @property
    def roof_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the roof components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the roof components.
        """
        return sorted(
            [c for c in self._components if c.is_roof], key=lambda _: _.display_name
        )

    @property
    def above_grade_wall_components(self) -> List[PhxComponentOpaque]:
        """Returns a sorted list (by display name) of all the above grade wall components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the above grade wall components.
        """
        return sorted(
            [c for c in self._components if c.is_above_grade_wall],
            key=lambda _: _.display_name,
        )

    @property
    def shading_components(self) -> List[PhxComponentOpaque]:
        """Returns a list of all the opaque shade components in the building.

        Returns:
        --------
            * (List[PhxComponentOpaque]) A sorted list of all the shading components.
        """
        return sorted(
            [c for c in self._components if c.is_shade], key=lambda _: _.display_name
        )

    @property
    def polygon_ids(self) -> Set[int]:
        """Return a Set of all the Polygon IDs of all Polygons from all the Components in the building."""
        p_ids = set()
        for compo in self.all_components:
            p_ids.update(compo.polygon_ids)
        return p_ids

    @property
    def polygons(self) -> List[geometry.PhxPolygon]:
        """Returns a list of all the Polygons of all the Components in the building."""
        return [poly for component in self.all_components for poly in component.polygons]

    def __bool__(self) -> bool:
        return bool(self.opaque_components) or bool(self.zones)

    def get_total_gross_wall_area(self) -> float:
        """Returns the total wall area of all the opaque components in the building."""
        return sum(
            c.get_total_gross_component_area() for c in self.above_grade_wall_components
        )

    def get_total_net_wall_area(self) -> float:
        """Returns the total net wall area of all the opaque components in the building."""
        return sum(
            c.get_total_net_component_area() for c in self.above_grade_wall_components
        )

    def get_total_gross_roof_area(self) -> float:
        """Returns the total roof area of all the opaque components in the building."""
        return sum(c.get_total_gross_component_area() for c in self.roof_components)

    def get_total_net_roof_area(self) -> float:
        """Returns the total net roof area of all the opaque components in the building."""
        return sum(c.get_total_net_component_area() for c in self.roof_components)

    def get_total_wall_aperture_area(self) -> float:
        """Returns the total window area of all the opaque components in the building."""
        return sum(c.get_total_aperture_area() for c in self.wall_aperture_components)

    def get_total_roof_aperture_area(self) -> float:
        """Returns the total skylight area of all the opaque components in the building."""
        return sum(c.get_total_aperture_area() for c in self.roof_aperture_components)

    def scale_all_wall_aperture_components(self, _scale_factor: float = 1.0) -> None:
        """Scale all the wall-aperture component's polygons by a given factor.

        Args:
        -----
            * scale_factor (float): The factor to scale the aperture components by.
        """
        for c in self.wall_aperture_components:
            c.scale(_scale_factor)

    def scale_all_roof_aperture_components(self, _scale_factor: float = 1.0) -> None:
        """Scale all the roof-aperture component's polygons by a given factor.

        Args:
        -----
            * scale_factor (float): The factor to scale the aperture components by.
        """
        for c in self.roof_aperture_components:
            c.scale(_scale_factor)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""PHX Project Classes"""

from __future__ import annotations
from typing import ClassVar, List, Dict, Optional, Any, Set
from dataclasses import dataclass, field

from PHX.model.building import PhxBuilding, PhxZone
from PHX.model.certification import PhxPhiusCertification, PhxPhiCertification
from PHX.model.constructions import PhxConstructionOpaque, PhxConstructionWindow
from PHX.model.geometry import PhxGraphics3D
from PHX.model.hvac.collection import PhxMechanicalSystemCollection
from PHX.model.phx_site import PhxSite
from PHX.model.schedules import ventilation, occupancy, lighting
from PHX.model.shades import PhxWindowShade
from PHX.model.utilization_patterns import (
    UtilizationPatternCollection_Occupancy,
    UtilizationPatternCollection_Ventilation,
    UtilizationPatternCollection_Lighting,
)


@dataclass
class WufiPlugin:
    insert_plugin: bool = False
    name_dll: Optional[Any] = None
    status_plugin: Optional[Any] = None


@dataclass
class PhxVariant:
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    name: Optional[str] = "unnamed_variant"
    remarks: Optional[str] = None
    plugin: Optional[WufiPlugin] = field(default_factory=WufiPlugin)
    building: PhxBuilding = field(default_factory=PhxBuilding)
    phius_cert: PhxPhiusCertification = field(default_factory=PhxPhiusCertification)
    phi_cert: PhxPhiCertification = field(default_factory=PhxPhiCertification)
    site: PhxSite = field(default_factory=PhxSite)
    mech_systems: PhxMechanicalSystemCollection = field(
        default_factory=PhxMechanicalSystemCollection
    )

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def graphics3D(self):
        """Collects all of the geometry (Polygons, Vertices) in the Project."""
        phx_graphics3D = PhxGraphics3D()
        for phx_component in self.building.all_components:
            phx_graphics3D.add_polygons(phx_component.polygons)
        return phx_graphics3D

    @property
    def phi_certification_major_version(self) -> int:
        """Return the PHI Certification Version Number."""
        return self.phi_cert.version

    @property
    def zones(self) -> List[PhxZone]:
        """Return a list of all the PHX Zones in the variant.building"""
        return self.building.zones

    def get_total_gross_wall_area(self) -> float:
        """Returns the total wall area of the variant.building"""
        return self.building.get_total_gross_wall_area()

    def get_total_net_wall_area(self) -> float:
        """Returns the total net wall area of the variant.building"""
        return self.building.get_total_net_wall_area()

    def get_total_wall_aperture_area(self) -> float:
        """Returns the total window area of the variant.building"""
        return self.building.get_total_wall_aperture_area()

    def get_total_gross_roof_area(self) -> float:
        """Returns the total wall area of the variant.building"""
        return self.building.get_total_gross_roof_area()

    def get_total_net_roof_area(self) -> float:
        """Returns the total net wall area of the variant.building"""
        return self.building.get_total_net_roof_area()

    def get_total_roof_aperture_area(self) -> float:
        """Returns the total window area of the variant.building"""
        return self.building.get_total_roof_aperture_area()


@dataclass
class ProjectData_Agent:
    name: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    post_code: Optional[str] = None
    telephone: Optional[str] = None
    email: Optional[str] = None
    license_number: Optional[str] = None


@dataclass
class PhxProjectData:
    customer: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    building: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    owner: ProjectData_Agent = field(default_factory=ProjectData_Agent)
    designer: ProjectData_Agent = field(default_factory=ProjectData_Agent)

    project_date: str = ""
    owner_is_client: bool = False
    year_constructed: int = 0
    image: Optional[bool] = None


@dataclass
class PhxProject:
    name: str = "unnamed_project"
    assembly_types: Dict[str, PhxConstructionOpaque] = field(default_factory=dict)
    window_types: Dict[str, PhxConstructionWindow] = field(default_factory=dict)
    shade_types: Dict[str, PhxWindowShade] = field(default_factory=dict)
    utilization_patterns_ventilation: UtilizationPatternCollection_Ventilation = field(
        default_factory=UtilizationPatternCollection_Ventilation
    )
    utilization_patterns_occupancy: UtilizationPatternCollection_Occupancy = field(
        default_factory=UtilizationPatternCollection_Occupancy
    )
    utilization_patterns_lighting: UtilizationPatternCollection_Lighting = field(
        default_factory=UtilizationPatternCollection_Lighting
    )
    variants: List[PhxVariant] = field(default_factory=list)
    project_data: PhxProjectData = field(default_factory=PhxProjectData)
    data_version: int = 48
    unit_system: int = 1  # SI
    program_version: str = "3.2.0.1"
    scope: int = 3
    visualized_geometry: int = 2

    def add_new_variant(self, _variant: PhxVariant) -> None:
        """Adds a new PHX Variant to the Project."""
        self.variants.append(_variant)

    @property
    def assembly_type_id_numbers(self) -> Set[int]:
        return {assembly.id_num for assembly in self.assembly_types.values()}

    @property
    def window_type_id_numbers(self) -> Set[int]:
        return {window.id_num for window in self.window_types.values()}

    @property
    def shade_type_id_numbers(self) -> Set[int]:
        return {shade.id_num for shade in self.shade_types.values()}

    def add_assembly_type(self, _assembly_type: PhxConstructionOpaque, _key=None) -> None:
        """Adds a new PhxConstructionOpaque to the Project's collection"""
        # -- be sure to the ID-Num
        if _assembly_type.id_num in self.assembly_type_id_numbers:
            _assembly_type.id_num = max(self.assembly_type_id_numbers) + 1

        if _key:
            self.assembly_types[_key] = _assembly_type
        else:
            self.assembly_types[_assembly_type.identifier] = _assembly_type

    def add_new_window_type(self, _window_type: PhxConstructionWindow, _key=None) -> None:
        """Adds a new PhxConstructionWindow to the Project's collection"""
        if _window_type.id_num in self.window_type_id_numbers:
            _window_type.id_num = max(self.window_type_id_numbers) + 1

        if _key:
            self.window_types[_key] = _window_type
        else:
            self.window_types[_window_type.identifier] = _window_type

    def get_window_type(self, _key: str) -> PhxConstructionWindow:
        """Returns the PhxConstructionWindow with the specified key"""
        return self.window_types[_key]

    def add_new_shade_type(self, _shade_type: PhxWindowShade, _key=None) -> None:
        """Adds a new PhxWindowShade to the Project's collection"""
        if _shade_type.id_num in self.shade_type_id_numbers:
            _shade_type.id_num = max(self.shade_type_id_numbers) + 1

        if _key:
            self.shade_types[_key] = _shade_type
        else:
            self.shade_types[_shade_type.identifier] = _shade_type

    def vent_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Ventilation schedule collection already includes the specified key."""
        return self.utilization_patterns_ventilation.key_is_in_collection(_key)

    def occupancy_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Occupancy schedule collection already includes the specified key."""
        return self.utilization_patterns_occupancy.key_is_in_collection(_key)

    def lighting_sched_in_project_collection(self, _key: str) -> bool:
        """See if the project Lighting schedule collection already includes the specified key."""
        return self.utilization_patterns_lighting.key_is_in_collection(_key)

    def add_vent_sched_to_collection(
        self, vent_sched: ventilation.PhxScheduleVentilation
    ) -> None:
        """Add a new Ventilation schedule to the project's collection."""
        self.utilization_patterns_ventilation.add_new_util_pattern(vent_sched)

    def add_occupancy_sched_to_collection(
        self, vent_sched: Optional[occupancy.PhxScheduleOccupancy]
    ) -> None:
        """Add a new Occupancy schedule to the project's collection."""
        self.utilization_patterns_occupancy.add_new_util_pattern(vent_sched)

    def add_lighting_sched_to_collection(
        self, lighting_sched: Optional[lighting.PhxScheduleLighting]
    ) -> None:
        """Add a new Occupancy schedule to the project's collection."""
        self.utilization_patterns_lighting.add_new_util_pattern(lighting_sched)

    def get_total_gross_wall_area(self) -> float:
        """Get the total gross wall area for all variants in the project (ignoring any apertures)."""
        return sum([variant.get_total_gross_wall_area() for variant in self.variants])

    def get_total_net_wall_area(self) -> float:
        """Get the total net wall area for all variants in the project (ignoring any apertures)."""
        return sum([variant.get_total_net_wall_area() for variant in self.variants])

    def get_total_gross_roof_area(self) -> float:
        """Get the total gross roof area for all variants in the project."""
        return sum([variant.get_total_gross_roof_area() for variant in self.variants])

    def get_total_net_roof_area(self) -> float:
        """Get the total net roof area for all variants in the project."""
        return sum([variant.get_total_net_roof_area() for variant in self.variants])

    def get_total_wall_aperture_area(self) -> float:
        """Get the total window area for all variants in the project."""
        return sum([variant.get_total_wall_aperture_area() for variant in self.variants])

    def get_total_roof_aperture_area(self) -> float:
        """Get the total roof aperture area for all variants in the project."""
        return sum([variant.get_total_roof_aperture_area() for variant in self.variants])

    def __str__(self):
        return f"{self.__class__.__name__}"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "Areas" worksheet."""

from __future__ import annotations
from typing import List, Optional, Dict, Generator, Tuple

from ph_units.unit_type import Unit

from PHX.xl import xl_app, xl_data
from PHX.xl.xl_data import col_offset
from PHX.PHPP.phpp_model import areas_surface, areas_data, areas_thermal_bridges
from PHX.PHPP.phpp_localization import shape_model


class AreasInputLocation:
    """Generic input item for Areas worksheet items."""

    def __init__(
        self,
        _xl: xl_app.XLConnection,
        _sheet_name: str,
        _search_col: str,
        _search_item: str,
        _input_row_offset: int,
    ):
        self.xl = _xl
        self.sheet_name = _sheet_name
        self.search_col = _search_col
        self.search_item = _search_item
        self.input_row_offset = _input_row_offset

    def find_input_row(self, _row_start: int = 1, _row_end: int = 200) -> int:
        """Return the row number where the search-item is found input."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.sheet_name,
            _col=self.search_col,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.search_item in str(val):
                return i + self.input_row_offset

        raise Exception(
            f'\n\tError: Not able to find the "{self.search_item}" input '
            f'section of the "{self.sheet_name}" worksheet? Please be sure '
            f'the item is note with the "{self.search_item}" flag in column {self.search_col}?'
        )


class Surfaces:
    def __init__(
        self,
        _xl: xl_app.XLConnection,
        _shape: shape_model.Areas,
        _group_type_exposures: Dict[int, str],
    ) -> None:
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None
        self._section_last_entry_row: Optional[int] = None
        self.surface_cache = {}
        self.group_type_exposures = _group_type_exposures

    @property
    def section_header_row(self) -> int:
        """Return the row number of the 'Area input' section header."""
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Area input' section."""
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    @property
    def section_last_entry_row(self) -> int:
        """Return the row number of the last user-input entry row in the 'Area input' section."""
        if not self._section_last_entry_row:
            self._section_last_entry_row = self.find_section_last_entry_row()
        return self._section_last_entry_row

    def find_section_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        """Return the row number of the 'Area input' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.surface_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=1):
            if val == self.shape.surface_rows.locator_string_header:
                return i

        raise Exception(
            f'\n\tError: Not able to find the "Areas input" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section begins '
            f'with the "{self.shape.surface_rows.locator_string_header}" flag in '
            f"column {self.shape.surface_rows.locator_col_header}."
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Area input' section."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.surface_rows.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # type: ignore - Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.surface_rows.locator_string_entry:
                return i

        raise Exception(
            f'\n\tError: Not able to find the first surface entry row in the "Areas input" section?'
        )

    def find_section_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the row number of the last user-input entry row in the 'Area input' section."""

        if not _start_row:
            _start_row = self.section_first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f'\n\tError: Not able to find the last surface entry row in the "Areas input" section?'
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.surface_rows.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            try:
                # -- See if the row has a number
                int(val)  # type: ignore
            except:
                # -- if not, that is the end of the input block
                return i - 1
        else:
            return self.find_section_last_entry_row(_row_end)

    def get_surface_phpp_id_by_name(self, _name: str, _use_cache: bool = False) -> str:
        """Return the PHPP-Style id ("1-NorthRoofSurface", ...) when given the surface name."""

        # -- Try and return the Cached value first.
        if _use_cache:
            try:
                return self.surface_cache[_name]
            except KeyError:
                pass

        row = self.xl.get_row_num_of_value_in_column(
            sheet_name=self.shape.name,
            row_start=self.section_first_entry_row,
            row_end=self.section_first_entry_row + 500,
            col=str(self.shape.surface_rows.inputs.description.column),
            find=_name,
        )

        if not row:
            raise Exception(
                f"Error: Cannot locate the phpp surface named: {_name} in"
                f"column {self.shape.surface_rows.inputs.description.column}?"
            )

        # -- Figure out the right full PHPP name, with the ID-number prefix
        prefix_range = f"{col_offset(str(self.shape.surface_rows.inputs.description.column), -1)}{row}"
        prefix_value = self.xl.get_data(self.shape.name, prefix_range)

        try:
            prefix_value = str(prefix_value)  # prefix_value comes in from excel as "1.0"
            prefix_value = float(prefix_value)  # Can't convert "1.0" -> 1 directly??
            prefix_value = int(prefix_value)
        except:
            msg = (
                f"\n\tError: Something went wrong trying to find the PHPP-ID "
                f"number for the surface: '{_name}'? Expected to find an integer "
                f"value but got: '{prefix_value}' at {self.shape.name}:{prefix_range}"
            )
            raise Exception(msg)

        self.xl.output(f"Getting PHPP Surface id for {_name}")
        name = f"{prefix_value}-{_name}"

        # -- Save in cache
        self.surface_cache[_name] = name

        return name

    def get_all_construction_names(self) -> set[str]:
        """Return a set of all the construction names used in the Areas worksheet."""
        assembly_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.surface_rows.inputs.assembly_id.column),
            self.section_first_entry_row,
            self.section_last_entry_row,
        )
        return {areas_surface.get_name_from_assembly_id(_d) for _d in assembly_data}

    @property
    def all_surface_rows(
        self,
    ) -> Generator[Tuple[int, areas_surface.ExistingSurfaceRow], None, None]:
        """Return a generator of all the row_nums and surface rows in the Areas worksheet.

        Yields:
        -------
            * Tuple[int, ExistingSurfaceRow]: The row-number and the surface row object.
        """

        for i in range(self.section_first_entry_row, self.section_last_entry_row):
            existing_surface_row = areas_surface.ExistingSurfaceRow(
                self.shape,
                self.xl.get_single_row_data(self.shape.name, i),
                self.group_type_exposures,
            )
            yield i, existing_surface_row


class ThermalBridges:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Areas) -> None:
        self.xl = _xl
        self.shape = _shape
        self._section_header_row: Optional[int] = None
        self._section_first_entry_row: Optional[int] = None

    @property
    def section_header_row(self) -> int:
        if not self._section_header_row:
            self._section_header_row = self.find_section_header_row()
        return self._section_header_row

    @property
    def section_first_entry_row(self) -> int:
        if not self._section_first_entry_row:
            self._section_first_entry_row = self.find_section_first_entry_row()
        return self._section_first_entry_row

    def find_section_header_row(self, _row_start: int = 100, _row_end: int = 500) -> int:
        """Return the row number of the 'Thermal Bridge input' section header."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.thermal_bridge_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if val == self.shape.thermal_bridge_rows.locator_string_header:
                return i

        # -- If isn't found in the first batch of data, recursively look further.
        if _row_end < 10_000:
            return self.find_section_header_row(
                _row_start=_row_end,
                _row_end=_row_end + 1_000,
            )

        raise Exception(
            f'\n\tError: Not able to find the "Thermal Bridge input" input section of '
            f'the "{self.shape.name}" worksheet? Please be sure the section begins '
            f'with the "{self.shape.thermal_bridge_rows.locator_string_header}" flag in '
            f"column {self.shape.thermal_bridge_rows.locator_col_header}."
        )

    def find_section_first_entry_row(self) -> int:
        """Return the row number of the very first user-input entry row in the 'Thermal Bridge input' section."""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.thermal_bridge_rows.locator_col_entry,
            _row_start=self.section_header_row,
            _row_end=self.section_header_row + 25,
        )

        for i, val in enumerate(xl_data, start=self.section_header_row):
            try:
                val = str(int(val))  # type: ignore - Value comes in as  "1.0" from Excel?
            except:
                continue

            if val == self.shape.thermal_bridge_rows.locator_string_entry:
                return i

        raise Exception(
            f'\n\tError: Not able to find the first Thermal Bridge entry row in the "Thermal Bridge input" section?'
        )


class Areas:
    """IO Controller for the PHPP Areas worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Areas) -> None:
        self.xl = _xl
        self.shape = _shape
        self.group_type_exposures = self.get_group_type_exposures()
        self.surfaces = Surfaces(self.xl, self.shape, self.group_type_exposures)
        self.thermal_bridges = ThermalBridges(self.xl, self.shape)

    def write_thermal_bridges(
        self, _tbs: List[areas_thermal_bridges.ThermalBridgeRow]
    ) -> None:
        """Write all of the the thermal bridge data to the PHPP Areas worksheet."""

        for i, tb in enumerate(_tbs, start=self.thermal_bridges.section_first_entry_row):
            for item in tb.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def write_surfaces(self, _surfaces: List[areas_surface.SurfaceRow]) -> None:
        """Write all of the the surface data to the PHPP Areas worksheet."""
        start = self.surfaces.section_first_entry_row
        for i, surface in enumerate(_surfaces, start=start):
            for item in surface.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)

    def _create_input_location_object(
        self, _phpp_model_obj: areas_data.AreasInput
    ) -> AreasInputLocation:
        """Create and setup the AreasInputLocation object with the correct data."""
        phpp_obj_shape: shape_model.AreasDataInput = getattr(
            self.shape, _phpp_model_obj.input_type
        )
        return AreasInputLocation(
            _xl=self.xl,
            _sheet_name=self.shape.name,
            _search_col=phpp_obj_shape.locator_col,
            _search_item=phpp_obj_shape.locator_string,
            _input_row_offset=phpp_obj_shape.input_row_offset,
        )

    def write_item(self, _phpp_model_obj: areas_data.AreasInput) -> None:
        """Write the VerificationInputItem item out to the PHPP Areas Worksheet."""
        input_object = self._create_input_location_object(_phpp_model_obj)
        input_row = input_object.find_input_row()
        xl_item = _phpp_model_obj.create_xl_item(self.shape.name, input_row)
        self.xl.write_xl_item(xl_item)

    def get_group_type_exposures(self) -> Dict[int, str]:
        """Return the group type exposures dictionary from the PHPP Areas worksheet."""
        zone_type_letters = self.xl.get_single_column_data(
            self.shape.name,
            self.shape.summary_rows.temp_zones,
            1,
            30,
        )
        zone_group_numbers = self.xl.get_single_column_data(
            self.shape.name,
            self.shape.summary_rows.group_number,
            1,
            30,
        )

        d = {}
        for group_num, temp_zone in zip(zone_group_numbers, zone_type_letters):
            if not group_num or not temp_zone:
                continue

            try:
                d[int(group_num)] = temp_zone
            except TypeError:
                pass
            except ValueError:
                pass
        return d

    def get_total_net_wall_area(self) -> Unit:
        """Return the total net (apertures punched) wall area from the PHPP Areas worksheet."""
        item = self.shape.defined_ranges.exposed_wall_area
        value = self.xl.get_single_data_item(self.shape.name, f"{item.column}{item.row}")
        return Unit(float(value or 0.0), str(item.unit))

    def get_total_net_roof_area(self) -> Unit:
        """Return the total net (apertures punched) Roof area from the PHPP Areas worksheet."""
        item = self.shape.defined_ranges.roof_ceiling_area
        value = self.xl.get_single_data_item(self.shape.name, f"{item.column}{item.row}")
        return Unit(float(value or 0.0), str(item.unit))

    def get_total_vertical_window_area(self) -> Unit:
        """Return the total window area from the PHPP Areas worksheet.


        Note: this value includes any non-horizontal windows and so may include "skylights",
        if the skylights are on sloped roof surfaces. Use the
        io_windows.get_total_window_area() function to get the total window area only.
        """
        unit_type = self.shape.defined_ranges.window_area_north.unit
        start = self.shape.defined_ranges.window_area_north.xl_range
        end = self.shape.defined_ranges.window_area_west.xl_range
        _range = f"{start}:{end}"

        xl_data: List[float] = self.xl.get_data(self.shape.name, _range)  # type: ignore
        unit_data: List[Unit] = [Unit(v, str(unit_type)) for v in xl_data]

        return sum(unit_data) or Unit(0.0, str(unit_type))

    def get_total_horizontal_window_area(self) -> Unit:
        """Return the total skylight area from the PHPP Areas worksheet.

        Note: this value includes ONLY flat windows and not all the "skylights",
        if the skylights are on sloped roof surfaces. Use the
        io_windows.get_total_skylight_area() function to get the total skylight area.
        """
        item = self.shape.defined_ranges.window_area_horizontal
        value = self.xl.get_single_data_item(self.shape.name, f"{item.column}{item.row}")
        return Unit(float(value or 0.0), str(item.unit))

    def set_surface_row_construction(
        self, _row_num: int, _phpp_constriction_id: str
    ) -> None:
        """Set the construction-id for the surface row in the PHPP Areas worksheet."""
        col = self.shape.surface_rows.inputs.assembly_id.column
        item = xl_data.XlItem(self.shape.name, f"{col}{_row_num}", _phpp_constriction_id)
        self.xl.write_xl_item(item)

    def set_surface_row_solar_absorptivity(
        self, _row_num: int, _absorptivity: float = 0.75
    ) -> None:
        """Set the solar absorptivity for the surface row in the PHPP Areas worksheet."""
        col = self.shape.surface_rows.inputs.absorptivity.column
        item = xl_data.XlItem(self.shape.name, f"{col}{_row_num}", _absorptivity)
        self.xl.write_xl_item(item)

    def set_surface_row_emissivity(
        self, _row_num: int, _emissivity: float = 0.90
    ) -> None:
        """Set the emissivity for the surface row in the PHPP Areas worksheet."""
        col = self.shape.surface_rows.inputs.emissivity.column
        item = xl_data.XlItem(self.shape.name, f"{col}{_row_num}", _emissivity)
        self.xl.write_xl_item(item)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP "SolarPV" worksheet."""

from __future__ import annotations
from dataclasses import dataclass, field
from functools import partial
from typing import Dict, Any, List

from ph_units.unit_type import Unit

from PHX.xl import xl_app
from PHX.PHPP.phpp_localization import shape_model


@dataclass
class SolarPVData:
    """Convenience class for organizing and cleaning the data."""

    display_name: str = ""
    footprint: Unit = field(default_factory=Unit)
    size: Unit = field(default_factory=Unit)
    annual_pv_energy: Unit = field(default_factory=Unit)

    @classmethod
    def from_PHPP_data(cls, _data: Dict[str, Any]) -> SolarPVData:
        """Clean up the data coming in from PHPP"""
        obj = cls()

        obj.display_name = _data["display_name"]
        obj.footprint = Unit(_data["footprint"], _data["footprint_unit"])
        obj.size = Unit(_data["size"], "KW")
        obj.annual_pv_energy = Unit(_data["pv_energy"], _data["energy_unit"])

        return obj


class SolarPV:
    """IO Controller for the PHPP 'Solar PV' worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.SolarPv) -> None:
        self.xl = _xl
        self.shape = _shape

    def get_phpp_data(self) -> List[SolarPVData]:
        """Get the data from the PHPP worksheet."""
        start = f"{self.shape.columns.systems_start}{self.shape.rows.systems_start}"
        end = f"{self.shape.columns.systems_end}{self.shape.rows.systems_end}"
        phpp_data = self.xl.get_data_by_columns(self.shape.name, f"{start}:{end}")

        pv_systems = []
        for i, sys_data in enumerate(phpp_data, start=self.shape.rows.systems_start):
            try:
                current = float(sys_data[self.shape.rows.current - i])  # type: ignore
                voltage = float(sys_data[self.shape.rows.voltage - i])  # type: ignore
                num_panes = float(sys_data[self.shape.rows.num_panels - i])  # type: ignore
                calculated_size_kw = (current * voltage * num_panes) / 1000
            except:
                calculated_size_kw = 0.0

            pv_sys = SolarPVData.from_PHPP_data(
                {
                    "footprint_unit": self.shape.footprint_unit,
                    "energy_unit": self.shape.energy_unit,
                    "display_name": sys_data[self.shape.rows.name - i],
                    "footprint": sys_data[self.shape.rows.footprint - i],
                    "size": calculated_size_kw,
                    "pv_energy": sys_data[self.shape.rows.annual_energy - i],
                }
            )

            pv_systems.append(pv_sys)

        return pv_systems



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to build PHX 'RoomVentilation' entities from Honeybee-PH Spaces"""

from honeybee import room
from honeybee_energy.properties.room import RoomEnergyProperties

from honeybee_ph import space
from honeybee_ph.properties.room import RoomPhProperties
from honeybee_ph.properties.space import SpacePhProperties

from honeybee_energy_ph.properties.hvac.idealair import IdealAirSystemPhProperties
from honeybee_energy_ph.properties.ruleset import ScheduleRulesetPhProperties

from honeybee_ph_utils.occupancy import hb_room_ppl_per_area
from honeybee_ph_utils.ventilation import hb_room_vent_flowrates

from PHX.model import spaces
from PHX.model.utilization_patterns import (
    UtilizationPatternCollection_Ventilation,
    UtilizationPatternCollection_Occupancy,
    UtilizationPatternCollection_Lighting,
)


def calc_space_ventilation_flow_rate(_space: space.Space) -> float:
    """Calculate and return the total peak ventilation flow rate for a Space.

    This function will determine the peak flow by-person, by-area, by-zone, and by-ach
    and return the sum of all four flow-rate types.

    Arguments:
    ----------
        * _space (space.Space): The Honeybee-PH Space to use as the source.

    Returns:
    --------
        * (float)
    """
    # Type Aliases
    host_room: room.Room = _space.host  # type: ignore
    host_room_prop_ph: RoomPhProperties = host_room.properties.ph  # type: ignore

    (
        flow_per_person,
        flow_per_area,
        air_changes_per_hour,
        flow_per_zone,
    ) = hb_room_vent_flowrates(host_room)
    # TODO: Unweighted or weighted? Which is right?
    ref_flr_area = _space.floor_area

    # -- Basic flow rates
    m3s_by_occupancy = ref_flr_area * hb_room_ppl_per_area(host_room) * flow_per_person
    m3s_by_area = ref_flr_area * flow_per_area

    # -- Figure out % of the HB-Room that the Space represents
    # -- For the Flow-by-Zone and Flow-by_ACH, need to calc the Room total flow
    # -- and then calc the % of that total that this one space represents.
    hb_room_total_space_fa = host_room_prop_ph.total_space_floor_area
    space_percent_of_total = ref_flr_area / hb_room_total_space_fa

    m3s_by_ach = (air_changes_per_hour * space_percent_of_total) / 3_600
    m3s_by_zone = flow_per_zone * space_percent_of_total

    return (m3s_by_occupancy + m3s_by_area + m3s_by_zone + m3s_by_ach) * 3_600


def create_room_from_space(
    _space: space.Space,
    _vent_sched_collection: UtilizationPatternCollection_Ventilation,
    _occ_sched_collection: UtilizationPatternCollection_Occupancy,
    _lighting_sched_collection: UtilizationPatternCollection_Lighting,
) -> spaces.PhxSpace:
    """Create a new RoomVentilation object with attributes based on a Honeybee-PH Space

    Arguments:
    ----------
        * _space (space.Space): The Honeybee-PH Space to use as the source.
        * _vent_sched_collection (UtilizationPatternCollection_Ventilation): A collection of
            ventilation schedules to use for the new PHX-Room.
        * _occ_sched_collection (UtilizationPatternCollection_Occupancy): A collection of
            occupancy schedules to use for the new PHX-Room.

    Returns:
    --------
        * (ventilation.RoomVentilation): The new PHX-Room with attributes based on the Honeybee Space.
    """
    # -- Type Aliases
    host_room_prop_energy: RoomEnergyProperties = _space.host.properties.energy  # type: ignore
    hbe_vent = host_room_prop_energy.ventilation
    hbe_vent_sched = hbe_vent.schedule
    hbe_vent_sched_prop_ph: ScheduleRulesetPhProperties = hbe_vent_sched.properties.ph  # type: ignore
    hbe_occ = host_room_prop_energy.people
    hbe_lighting = host_room_prop_energy.lighting
    space_prop_ph: SpacePhProperties = _space.properties.ph  # type: ignore
    hbe_hvac = host_room_prop_energy.hvac
    hbe_hvac_prop_ph: IdealAirSystemPhProperties = hbe_hvac.properties.ph  # type: ignore

    # -- Build the new PHX Space
    new_room = spaces.PhxSpace()

    new_room.display_name = _space.full_name
    new_room.wufi_type = _space.wufi_type
    new_room.quantity = _space.quantity
    new_room.floor_area = _space.floor_area
    new_room.weighted_floor_area = _space.weighted_floor_area
    new_room.clear_height = _space.avg_clear_height
    new_room.net_volume = _space.net_volume

    # -- Set the Room's Ventilation attributes
    new_room.ventilation.schedule = _vent_sched_collection[hbe_vent_sched.identifier]
    space_peak_flow_rate = calc_space_ventilation_flow_rate(_space)
    new_room.ventilation.load.flow_supply = space_peak_flow_rate
    new_room.ventilation.load.flow_extract = space_peak_flow_rate

    # -- TODO: FIX THIS TO A BETTER TECHNIQUE SOMEDAY, OVERRIDE WITH LOCAL INFO, IF ANY
    if space_prop_ph._v_sup is not None:
        new_room.ventilation.load.flow_supply = space_prop_ph._v_sup * 3_600

    if space_prop_ph._v_eta is not None:
        new_room.ventilation.load.flow_extract = space_prop_ph._v_eta * 3_600

    if space_prop_ph._v_tran is not None:
        new_room.ventilation.load.flow_transfer = space_prop_ph._v_tran * 3_600

    # -- Keep the Ventilation Equipment ID-Numbers aligned
    if hbe_hvac_prop_ph.ventilation_system:
        new_room.vent_unit_id_num = hbe_hvac_prop_ph.ventilation_system.id_num

    # -- Keep the new room's Occupancy reference aligned with the HB-Room's
    if hbe_occ:
        occ_sched_id = hbe_occ.occupancy_schedule.identifier
        new_room.occupancy.schedule = _occ_sched_collection[occ_sched_id]

    # -- Keep the new room's Lighting reference aligned with the HB-Room's
    if hbe_lighting:
        lighting_sched_id = hbe_lighting.schedule.identifier
        new_room.lighting.schedule = _lighting_sched_collection[lighting_sched_id]
        new_room.lighting.load.installed_w_per_m2 = hbe_lighting.watts_per_area

    return new_room



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Controller Class for the PHPP PER Worksheet."""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Union, Tuple, Collection, Sequence

from ph_units.unit_type import Unit

from PHX.xl.xl_app import XLConnection
from PHX.PHPP.phpp_localization import shape_model as shp
from PHX.PHPP.sheet_io.io_exceptions import (
    FindSectionMarkerException,
    PerReferenceAreaException,
)


# -----------------------------------------------------------------------------


class BaseBlock:
    """Base class for all PER Data Blocks (Heating, Cooling, etc..)"""

    def __init__(
        self,
        _host: PER,
        _xl: XLConnection,
        _shape: shp.PerDataBlock,
    ):
        self.host = _host
        self.xl = _xl
        self.shape = _shape
        self._phpp_data: Optional[Dict[str, List[Union[str, Unit]]]] = None
        self._block_header_row: Optional[int] = None
        self._block_start_row: Optional[int] = None
        self._block_end_row: Optional[int] = None
        self._reference_area_address: Optional[str] = None
        self._reference_area: Optional[float] = None

    @property
    def block_start_row(self) -> int:
        """Return the row number for the first row of the data block."""
        if not self._block_start_row:
            self._block_start_row = self.find_block_start_row()
        return self._block_start_row

    @property
    def block_end_row(self) -> int:
        """Return the row number for the last row of the data block."""
        if not self._block_end_row:
            self._block_end_row = self.find_block_end_row()
        return self._block_end_row

    @property
    def phpp_data(self) -> Dict[str, List[Union[str, Unit]]]:
        """Return all the Block PHPP Data as a Dict of Lists of values.

        -> {
            'P': ['Electricity (HP compact unit)', 'Electricity (heat pump)', ...],
            'T': [0.0 (KBTU/FT2), 0.8733713455642654 (KBTU/FT2), ...],
            'X': [0.0 (KBTU/FT2), 2.27076549846709 (KBTU/FT2), ...],
            }
        """
        if not self._phpp_data:
            self._phpp_data = self.get_data()
        return self._phpp_data

    @property
    def reference_area_address(self) -> str:
        """Return the range address for the Data type's Reference Area (TFA/ Footprint)."""
        if not self._reference_area_address:
            self._reference_area_address = self.host.shape.addresses.tfa  # Default
        return self._reference_area_address

    @property
    def reference_area(self) -> float:
        """Return the Data type's Reference Area (TFA/ Footprint)."""
        if not self._reference_area:
            try:
                self._reference_area = float(
                    self.xl.get_single_data_item(
                        self.worksheet_name, self.reference_area_address
                    )  # type: ignore
                )
            except:
                raise PerReferenceAreaException(
                    self.worksheet_name, self.reference_area_address
                )

        return self._reference_area

    @property
    def block_header_row(self) -> int:
        """Return the Row number where the block 'Heading' is found."""
        if not self._block_header_row:
            self._block_header_row = self.find_block_header_row()
        return self._block_header_row

    @property
    def worksheet_name(self) -> str:
        """Return the Name of the PER worksheet."""
        return self.host.shape.name

    @property
    def locator_column(self) -> str:
        """Return the column letter for the 'locator' column."""
        return self.host.shape.locator_col

    @property
    def column_final_energy(self) -> str:
        """Return the column letter for the 'Final' (Site) Energy Demand."""
        return self.host.shape.columns.final_energy

    @property
    def column_pe_energy(self) -> str:
        """Return the column letter for the 'Primary' (Source) Energy Demand."""
        return self.host.shape.columns.pe_energy

    @property
    def column_co2_emissions(self) -> str:
        """Return the column letter for the 'CO2 Emissions'."""
        return self.host.shape.columns.co2_emissions

    def find_block_header_row(self) -> int:
        """Locate the row number for the block heading"""
        data = self.xl.get_single_column_data(
            self.worksheet_name,
            self.locator_column,
            self.host.heading_row,
            self.host.heading_row + 100,
        )
        search_string = self.shape.locator_string_heading
        row_number = self.xl.find_row(search_string, data, self.host.heading_row)
        if not row_number:
            raise FindSectionMarkerException(
                search_string, self.worksheet_name, self.locator_column
            )
        return row_number

    def find_block_start_row(self) -> int:
        """Locate the row number for the first row of the data block."""
        data = self.xl.get_single_column_data(
            self.worksheet_name,
            self.locator_column,
            self.block_header_row,
            self.block_header_row + 10,
        )

        search_string = self.shape.locator_string_start
        row_number = self.xl.find_row(search_string, data, self.block_header_row)
        if not row_number:
            raise FindSectionMarkerException(
                search_string, self.worksheet_name, self.locator_column
            )
        return row_number

    def find_block_end_row(self) -> int:
        """Locate the row number for the last row of the data block."""
        data = self.xl.get_single_column_data(
            self.worksheet_name,
            self.locator_column,
            self.block_start_row,
            self.block_start_row + 25,
        )

        search_string = self.shape.locator_string_end
        row_number = self.xl.find_row(search_string, data, self.block_start_row)
        if not row_number:
            raise FindSectionMarkerException(
                search_string, self.worksheet_name, self.locator_column
            )
        return row_number

    def get_data(self) -> Dict[str, List[Union[str, Unit]]]:
        """Return all the Block's data from Excel"""
        start = f"{self.locator_column}{self.block_start_row}"
        end = f"{self.column_co2_emissions}{self.block_end_row}"
        address = f"{start}:{end}"
        xl_raw_data = self.xl.get_data_with_column_letters(self.worksheet_name, address)
        """
        xl_raw_data = {
            'P': ['Electricity (HP compact unit)', 'Electricity (heat pump)', ...], 
            'Q': [12.3, 1.9, 45.6, ...],
            ...
            'Z': [0.0, 0.0, 0.0, ...],
        }
        """

        # ----------------------------------------------------------------------
        # -- convert the raw numeric values to Units
        unit_dict: Dict[str, List[Union[str, Unit]]] = {}

        # -- Grab the index column data first
        unit_dict[self.locator_column] = [
            str(_) for _ in xl_raw_data[self.locator_column]
        ]

        # -- Get the numeric values as units
        unit_dict[self.host.shape.columns.final_energy] = [
            Unit(_, self.host.shape.unit)
            for _ in xl_raw_data[self.host.shape.columns.final_energy]
        ]
        unit_dict[self.host.shape.columns.pe_energy] = [
            Unit(_, self.host.shape.unit)
            for _ in xl_raw_data[self.host.shape.columns.pe_energy]
        ]

        return unit_dict

    def get_final_energy_by_fuel_type(self) -> Dict[str, Unit]:
        """Return the Block's Final (Site) Energy as a dict of Unit values.

        -> {
            'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
            'Electricity (heat pump)': 0.8733713455642654 (KBTU/FT2),
            ...
            }
        """
        use_types: List[str] = self.phpp_data[self.locator_column]  # type: ignore
        site_energy: List[Unit] = self.phpp_data[self.column_final_energy]  # type: ignore
        return {str(use_type): value for use_type, value in zip(use_types, site_energy)}

    def get_primary_energy_by_fuel_type(self) -> Dict[str, Unit]:
        """Return the Block's Primary (Source)) Energy as a dict of values.

        -> {
            'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
            'Electricity (heat pump)': 0.8733713455642654 (KBTU/FT2),
            ...
            }
        """
        use_types: List[str] = self.phpp_data[self.locator_column]  # type: ignore
        source_energy: List[Unit] = self.phpp_data[self.column_pe_energy]  # type: ignore
        return {str(t): e for t, e in zip(use_types, source_energy)}


# -----------------------------------------------------------------------------


class Heating(BaseBlock):
    """Heating Data Block."""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.tfa


class Cooling(BaseBlock):
    """Cooling Energy"""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.tfa


class DHW(BaseBlock):
    """Hot-Water Energy"""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.tfa


class HouseholdElectric(BaseBlock):
    """Household Electric (Lighting, Appliances, etc)"""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.tfa


class AdditionalGas(BaseBlock):
    """Gas for cooking, dryers"""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.tfa


class EnergyGeneration(BaseBlock):
    """Solar, Wind Energy Generation"""

    def __init__(self, _host: PER, _xl: XLConnection, _shape: shp.PerDataBlock):
        super().__init__(_host, _xl, _shape)
        self._reference_area_address = self.host.shape.addresses.footprint


# -----------------------------------------------------------------------------


@dataclass
class HeatingDeviceUsage:
    """Convenience class for organizing and cleaning the data."""

    device_type_name: str = "-"
    device_heating_percentage: Unit = field(default_factory=Unit)
    device_dhw_percentage: Unit = field(default_factory=Unit)

    @classmethod
    def from_phpp_data_row(cls, row: Sequence) -> HeatingDeviceUsage:
        """Create a new instance from a row of data from PHPP."""
        obj = cls()

        if not row:
            return obj

        if row[0] not in ["-", "", "None", None]:
            obj.device_type_name = str(row[0])

        if row[3] not in ["-", "", "None", None]:
            obj.device_heating_percentage = Unit(row[3], "-")

        if row[4] not in ["-", "", "None", None]:
            obj.device_dhw_percentage = Unit(row[4], "-")

        return obj


class PER:
    """IO Controller for the PHPP 'PER' worksheet."""

    def __init__(self, _xl: XLConnection, _shape: shp.Per):
        self.xl = _xl
        self.shape = _shape
        self.heating = Heating(self, _xl, _shape.heating)
        self.cooling = Cooling(self, _xl, _shape.cooling)
        self.dhw = DHW(self, _xl, _shape.dhw)
        self.household_electric = HouseholdElectric(self, _xl, _shape.household_electric)
        self.additional_gas = AdditionalGas(self, _xl, _shape.additional_gas)
        self.energy_generation = EnergyGeneration(self, _xl, _shape.energy_generation)
        self._heating_device_data = None  # local cache

    @property
    def heading_row(self) -> int:
        return 15

    @property
    def heating_device_data(self) -> Tuple[HeatingDeviceUsage]:
        """Return a list of HeatingDeviceUsage objects."""
        if not self._heating_device_data:
            self._heating_device_data = self.get_heating_device_type_data()
        return self._heating_device_data

    def get_final_kWh_m2_by_use_type(
        self,
    ) -> Dict[str, Dict[str, Unit]]:
        """Return a Dict of all the Site (Final) Energy [kWh/m2] by use-type."""
        return {
            "heating": self.heating.get_final_energy_by_fuel_type(),
            "cooling": self.cooling.get_final_energy_by_fuel_type(),
            "dhw": self.dhw.get_final_energy_by_fuel_type(),
            "household_electric": self.household_electric.get_final_energy_by_fuel_type(),
            "additional_gas": self.additional_gas.get_final_energy_by_fuel_type(),
            "energy_generation": self.energy_generation.get_final_energy_by_fuel_type(),
        }

    def get_final_kWh_by_fuel_type(self) -> Dict[str, Dict[str, Unit]]:
        """Return a Dict of all the Site (Final) Energy [kWh | kBtu] by use and fuel-type.

        -> {
            'heating': {
                'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
                ...
            'cooling': {
                'Electricity cooling (HP)': 1.1129088978478676 (KBTU/FT2),
                ...
                },
            'dhw': {
                'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
                ...
                },
            'household_electric': {
                'User electricity (lighting, electrical devices, etc.)': 2.5064700097780253 (KBTU/FT2),
                ...
                },
            'energy_generation':{
                'PV electricity': 19.687188186419185 (KBTU/FT2),
                ...
                }
        }
        """
        kwh_m2_data = self.get_final_kWh_m2_by_use_type()
        d: Dict[str, Dict[str, Unit]] = {}
        for use_type_name, use_type_values_by_fuel in kwh_m2_data.items():
            d[use_type_name] = {}
            energy_type = getattr(self, use_type_name)
            area_m2: float = energy_type.reference_area  # TFA or Footprint
            for fuel_type_name, fuel_type_values in use_type_values_by_fuel.items():
                try:
                    d[use_type_name][fuel_type_name] = fuel_type_values * area_m2
                except:
                    d[use_type_name][fuel_type_name] = fuel_type_values
        return d

    def get_primary_kWh_m2_by_use_type(self) -> Dict[str, Dict[str, Unit]]:
        """Return a Dict of all the Source (Primary) Energy [kWh/m2] by use-type."""
        self.xl.get_sheet_by_name(self.shape.name).activate()
        return {
            "heating": self.heating.get_primary_energy_by_fuel_type(),
            "cooling": self.cooling.get_primary_energy_by_fuel_type(),
            "dhw": self.dhw.get_primary_energy_by_fuel_type(),
            "household_electric": self.household_electric.get_primary_energy_by_fuel_type(),
            "additional_gas": self.additional_gas.get_primary_energy_by_fuel_type(),
            "energy_generation": self.energy_generation.get_primary_energy_by_fuel_type(),
        }

    def get_primary_kWh_by_fuel_type(self) -> Dict[str, Dict[str, Unit]]:
        """Return a Dict of all the Primary (Source) Energy [kWh | kBtu] by use and fuel type

        -> {
            'heating': {
                'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
                ...
            'cooling': {
                'Electricity cooling (HP)': 1.1129088978478676 (KBTU/FT2),
                ...
                },
            'dhw': {
                'Electricity (HP compact unit)': 0.0 (KBTU/FT2),
                ...
                },
            'household_electric': {
                'User electricity (lighting, electrical devices, etc.)': 2.5064700097780253 (KBTU/FT2),
                ...
                },
            'energy_generation':{
                'PV electricity': 19.687188186419185 (KBTU/FT2),
                ...
                }
        }
        """
        kwh_m2_data = self.get_primary_kWh_m2_by_use_type()
        d: Dict[str, Dict[str, Unit]] = {}
        for use_type_name, use_type_values_by_fuel in kwh_m2_data.items():
            d[use_type_name] = {}
            energy_type = getattr(self, use_type_name)
            area_m2: float = energy_type.reference_area  # TFA or Footprint
            for fuel_type_name, fuel_type_values in use_type_values_by_fuel.items():
                try:
                    d[use_type_name][fuel_type_name] = fuel_type_values * area_m2
                except:
                    d[use_type_name][fuel_type_name] = fuel_type_values
        return d

    def get_heating_device_type_data(self) -> Tuple[HeatingDeviceUsage]:
        """Return a Tuple of HeatingDeviceUsage objects from the PER worksheet."""
        data = self.xl.get_data(
            self.shape.name,
            f"{self.shape.heating_types.range_start}:{self.shape.heating_types.range_end}",
        )

        if not isinstance(data, Collection):
            raise ValueError(
                "There was a problem getting the heating-generation "
                f"type data from PHPP worksheet {self.shape.name}?"
            )

        return tuple(HeatingDeviceUsage.from_phpp_data_row(row) for row in data)
